
from . import local
import signal
import logging
import errno
import time
import os
import sys
from texttestlib import plugins
from texttestlib.utils import getPortListenErrorCode, getUserName
from threading import Thread, Lock
from queue import Queue
from fnmatch import fnmatch


class Ec2Machine:
    instanceTypeInfo = {"8xlarge": 32, "4xlarge": 16, "2xlarge": 8, "xlarge": 4, "large": 2, "medium": 1}

    def __init__(self, inst, synchDirs, app, subprocessLock, alreadyRunning):
        self.id = inst.id
        self.ip = inst.private_ip_address
        self.fullMachine = "ec2-user@" + self.ip
        self.cores = self.instanceTypeInfo.get(inst.instance_type.split(".")[-1], 1)
        self.synchDirs = synchDirs
        self.synchProc = None
        self.app = app
        self.remoteProcessInfo = {}
        self.remoteProcessInfoLock = Lock()
        self.thread = Thread(target=self.runThread)
        self.thread.setName("Machine_" + self.ip)
        self.startWaitCounter = 0
        self.diag = logging.getLogger("Ec2Machine")
        self.queue = Queue()
        self.errorMessage = ""
        self.subprocessLock = subprocessLock
        self.startMethod = None if alreadyRunning else inst.start

    def getNextJobId(self):
        return "job" + str(len(self.remoteProcessInfo)) + "_" + self.ip

    def getParents(self, dirs):
        parents = []
        for dir in dirs:
            parent = os.path.dirname(dir)
            if parent not in parents:
                parents.append(parent)
        return parents

    def isFull(self):
        return len(self.remoteProcessInfo) >= self.cores

    def hasJob(self, jobId):
        return jobId in self.remoteProcessInfo

    def setLocalProcessId(self, jobId, localPid):
        with self.remoteProcessInfoLock:
            remotePid = None
            if jobId in self.remoteProcessInfo:
                _, remotePid = self.remoteProcessInfo[jobId]
            self.remoteProcessInfo[jobId] = localPid, remotePid
        self.diag.info("Job ID " + jobId + " now got local PID " + localPid)

    def setRemoteProcessId(self, jobId, remotePid):
        with self.remoteProcessInfoLock:
            localPid, _ = self.remoteProcessInfo[jobId]
            self.remoteProcessInfo[jobId] = localPid, remotePid
        self.diag.info("Job ID " + jobId + " now got remote PID " + remotePid)

    def synchronise(self):
        parents = self.getParents(self.synchDirs)
        self.app.ensureRemoteDirExists(self.fullMachine, *parents)
        for dir in self.synchDirs:
            if not self.errorMessage:
                self.synchronisePath(dir)

    def synchronisePath(self, path):
        dirName = os.path.dirname(path)
        for _ in range(5):
            with self.subprocessLock:
                self.synchProc = self.app.getRemoteCopyFileProcess(path, "localhost", dirName, self.fullMachine)
            errorCode = self.synchProc.wait()
            if errorCode == 0:
                self.synchProc = None
                return
            else:
                time.sleep(1)

    def waitForStart(self):
        timeout = 1000
        self.diag.info("Waiting for response to ssh...")
        while self.startWaitCounter < timeout:
            ret = getPortListenErrorCode(self.ip, 22)
            if ret == 0 or self.errorMessage:
                self.startWaitCounter = 0
                break
            self.startWaitCounter += 1
            timedout = ret in [errno.EWOULDBLOCK, errno.ETIMEDOUT]
            if not timedout:
                time.sleep(1)

    def runThread(self):
        if self.startMethod:
            self.startMethod()  # should be self.waitForStart that is called here, not instance.start. Don't use boto methods in a thread!

        if self.errorMessage:
            return
        try:
            self.diag.info("Synchronising files with EC2 instance with private IP address '" + self.ip + "'...")
            self.synchronise()
        except plugins.TextTestError as e:
            self.errorMessage = "Failed to synchronise files with EC2 instance with private IP address '" + self.ip + "'\n" + \
                "Intended usage is to start an ssh-agent, and add the keypair for this instance to it, in your shell before starting TextTest from it.\n\n(" + str(
                    e) + ")\n"

        if self.errorMessage:
            return

        while True:
            self.diag.info("Waiting for new job for IP '" + self.ip + "'...")
            jobId, submitCallable = self.queue.get()
            if jobId is None:
                self.diag.info("No more tests for IP '" + self.ip + "', exiting.")
                return
            self.diag.info("Got job with ID " + jobId)
            localPid = self.doSubmit(submitCallable)
            self.setLocalProcessId(jobId, localPid)

    def doSubmit(self, submitCallable):
        with self.subprocessLock:
            localPid, _ = submitCallable()
            return localPid

    def cleanup(self, processes):
        # Return whether we are still using the machine in some way
        # i.e. if our thread is running or any of our processes are
        if self.thread.is_alive():
            self.queue.put((None, None))
            return True

        for localPid, _ in list(self.remoteProcessInfo.values()):
            if localPid in processes:
                proc = processes.get(localPid)
                if proc.poll() is None:
                    return True
        return False

    def getCommandArgsWithEnvironment(self, cmdArgs, slaveEnv):
        return [envVar + "=" + value for (envVar, value) in list(slaveEnv.items())] + cmdArgs

    def submitSlave(self, submitter, cmdArgs, slaveEnv, *args):
        jobId = self.getNextJobId()
        self.remoteProcessInfo[jobId] = None, None
        if not self.thread.is_alive():
            if self.startMethod:
                self.diag.info("Starting EC2 instance with private IP address '" + self.ip + "'...")
                try:
                    self.startMethod()
                except Exception as e:
                    sys.stderr.write("WARNING: failed to start instance with private IP address '" +
                                     self.ip + "'\n" + str(e))
                    return
                self.startMethod = self.waitForStart

            self.thread.start()
        argsWithEnv = self.getCommandArgsWithEnvironment(cmdArgs, slaveEnv)
        remoteCmdArgs = self.app.getCommandArgsOn(self.fullMachine, argsWithEnv, agentForwarding=True)
        self.queue.put((jobId, plugins.Callable(submitter, remoteCmdArgs, slaveEnv, *args)))
        return jobId

    def killRemoteProcess(self, jobId, sig):
        if self.synchProc:
            self.errorMessage = "Terminated test during file synchronisation"
            self.synchProc.send_signal(signal.SIGTERM)
            return True, None
        if self.startWaitCounter:
            self.errorMessage = "Terminated test while waiting for instance to start up"
            return True, None
        # ssh doesn't forward signals to remote processes.
        # We need to find it ourselves and send it explicitly. Can assume python exists remotely, but not much else.
        localPid, remotePid = self.waitForRemoteProcessId(jobId)
        if remotePid:
            cmdArgs = ["python", "-c", "\"import os; os.kill(" + remotePid + ", " + str(sig) + ")\""]
            self.app.runCommandOn(self.fullMachine, cmdArgs)
            return True, localPid
        else:
            return False, localPid

    def waitForRemoteProcessId(self, jobId):
        for _ in range(10):
            localPid, remotePid = self.remoteProcessInfo[jobId]
            if remotePid:
                return localPid, remotePid
            # Remote process exists but has not yet told us its process ID. Wait a bit and try again.
            time.sleep(1)
        return None, None

    def collectJobStatus(self, jobStatus, procStatus):
        if not self.errorMessage:
            for jobId, (localPid, _) in list(self.remoteProcessInfo.items()):
                if localPid:
                    if localPid in procStatus:
                        jobStatus[jobId] = procStatus[localPid]
                else:
                    jobStatus[jobId] = "SYNCH", "Synchronizing data with " + self.fullMachine


class QueueSystem(local.QueueSystem):
    userTagName = "TextTest user"

    def __init__(self, app):
        local.QueueSystem.__init__(self)
        self.nextMachineIndex = 0
        self.app = app
        self.subprocessLock = Lock()
        instances, runningIds = self.findInstances()
        synchDirs = self.getDirectoriesForSynch()
        self.machines = [Ec2Machine(inst, synchDirs, app, self.subprocessLock,
                                    inst.id in runningIds) for inst in instances]
        self.releasedMachines = []
        self.capacity = self.calculateCapacity()

    def calculateCapacity(self):
        return sum((m.cores for m in self.machines))

    def makeEc2Connection(self):
        import boto.ec2
        region = boto.ec2.connection.EC2Connection.DefaultRegionName  # stick to single region for now
        return boto.ec2.connect_to_region(region)

    def makeCloudwatchConnection(self):
        import boto.ec2.cloudwatch
        region = boto.ec2.connection.EC2Connection.DefaultRegionName  # stick to single region for now
        return boto.ec2.cloudwatch.connect_to_region(region)

    def getCores(self, inst, defValue=0):
        instanceSize = inst.instance_type.split(".")[-1]
        return Ec2Machine.instanceTypeInfo.get(instanceSize, defValue)

    def findInstances(self):
        if not self.app.getConfigValue("remote_copy_program"):
            sys.stderr.write(
                "Cannot run tests in EC2 cloud. You need to set 'remote_copy_program' in your config file to a program such as 'rsync'.\n")
            return [], []
        try:
            conn = self.makeEc2Connection()
        except ImportError:
            sys.stderr.write(
                "Cannot run tests in EC2 cloud. You need to install Python's boto package for this to work.\n")
            return [], []
        except:
            sys.stderr.write(
                "Failed to establish a connection to the EC2 cloud. Make sure your credentials are available in your .boto file.\n")
            return [], []
        instanceTags = self.app.getConfigValue("queue_system_resource")
        if "R" in self.app.inputOptions:
            instanceTags.append(self.app.inputOptions["R"])
        instances = self.findTaggedInstances(conn, instanceTags)
        if instances:
            running = self.getRunningIds(conn, instances)

            def getSortKey(inst):
                isRunning = inst.id in running
                cores = self.getCores(inst)
                return not isRunning, -cores, inst.private_ip_address

            instances.sort(key=getSortKey)
            maxCapacity = self.app.getConfigValue("queue_system_max_capacity")
            freeInstances, otherOwners = self.takeOwnership(conn, instances, maxCapacity)
            if freeInstances:
                self.disableAlarmActions(self.getAlarmNames(freeInstances))
            else:
                sys.stderr.write("Cannot run tests in EC2 cloud. " + str(len(instances)) + " running instances were found matching '" +
                                 ",".join(instanceTags) + "' in their tags, \nbut all are currently being used by the following users:\n" +
                                 "\n".join(otherOwners) + "\n\n")
            return freeInstances, running
        else:
            sys.stderr.write("Cannot run tests in EC2 cloud. No instances were found matching '" +
                             ",".join(instanceTags) + "' in their tags.\n")
            return [], []

    def cleanup(self, final=False):
        if final:
            # Processes might not be quite terminated, so we just hardcode that we release everything anyway
            self.releaseOwnership(self.machines)
        else:
            unusedMachines, usedMachines = [], []
            for machine in self.machines:
                if machine.cleanup(self.processes):
                    usedMachines.append(machine)
                else:
                    unusedMachines.append(machine)
            self.releaseOwnership(unusedMachines)
            self.machines = usedMachines
            self.releasedMachines = unusedMachines
        return False  # Submission is not really complete, as it happens in threads

    def matchesTag(self, instanceTags, tagName, tagPattern):
        tagValueForInstance = instanceTags.get(tagName, "")
        return fnmatch(tagValueForInstance, tagPattern)

    def parseTag(self, tag):
        return tag.split("=", 1) if "=" in tag else [tag, "1"]

    def findTaggedInstances(self, conn, instanceTags):
        instances = []
        parsedTags = [self.parseTag(tag) for tag in instanceTags]
        for inst in conn.get_only_instances():
            if inst.private_ip_address is not None and \
                    all((self.matchesTag(inst.tags, tagName, tagPattern) for tagName, tagPattern in parsedTags)):
                instances.append(inst)
        return instances

    def getRunningIds(self, conn, instances):
        ids = [inst.id for inst in instances]
        running = []
        for stat in conn.get_all_instance_status(ids):
            if stat.instance_status.status in ["ok", "initializing"]:
                running.append(stat.id)

        return running

    def tryAddTag(self, conn, instances, maxCapacity, myTag, otherOwners):
        # inst.tags is only a local cache. Try to avoid race conditions by getting the most up-to-date info possible.
        instanceIds = [instance.id for instance in instances]
        idsInUse = set()
        for inst in conn.get_only_instances(instance_ids=instanceIds):
            owner = inst.tags.get(self.userTagName, "")
            if owner:
                otherOwners.add(owner.split("_")[0])
                idsInUse.add(inst.id)

        tryOwnInstances, fallbackInstances = [], []
        capacity = 0
        for inst in instances:
            if inst.id not in idsInUse:
                if capacity < maxCapacity:
                    tryOwnInstances.append(inst.id)
                    inst.add_tag(self.userTagName, myTag)
                else:
                    fallbackInstances.append(inst)
                cores = self.getCores(inst, 1)
                capacity += cores

        return tryOwnInstances, fallbackInstances

    def getAlarmNames(self, instances):
        return ["stop-" + self.getInstanceName(iId) for iId in instances]

    def getInstanceName(self, instance):
        return instance if isinstance(instance, str) else instance.id

    def enableAlarmActions(self, alarmNames):
        conn = self.makeCloudwatchConnection()
        try:
            conn.enable_alarm_actions(alarmNames)
        except:
            pass

    def disableAlarmActions(self, alarmNames):
        conn = self.makeCloudwatchConnection()
        try:
            conn.disable_alarm_actions(alarmNames)
        except:
            plugins.printWarning("Could not disable CloudWatch alarms. Your user does not have permission for this.\n" +
                                 "The risk is that an instance will be shut down while you are using it, so it is suggested your request this permission from your administrator.\n")

    def takeOwnership(self, conn, instances, maxCapacity):
        myTag = getUserName() + "_" + plugins.startTimeString()
        otherOwners = set()
        tryOwnInstances, fallbackInstances = self.tryAddTag(conn, instances, maxCapacity, myTag, otherOwners)

        if not tryOwnInstances:
            return [], sorted(otherOwners)

        currTryInstances = tryOwnInstances
        ownInstances = []
        lostCapacity = 0
        time.sleep(0.5)  # add and check too close together makes racing more likely
        for _ in range(20):
            newInsts = conn.get_only_instances(instance_ids=currTryInstances)
            currTryInstances = []
            for inst in newInsts:
                owner = inst.tags.get(self.userTagName, "")
                if owner == myTag:
                    ownInstances.append(inst)
                elif owner:
                    # There's a race condition, somebody else grabbed it first, we drop it
                    otherOwners.add(owner.split("_")[0])
                    lostCapacity += self.getCores(inst, 1)
                else:
                    currTryInstances.append(inst.id)
            if currTryInstances:
                time.sleep(0.1)
            else:
                break

        def getOrigOrder(inst):
            return tryOwnInstances.index(inst.id)
        ownInstances.sort(key=getOrigOrder)

        if lostCapacity:
            fallbackInstances, fallbackOwners = self.takeOwnership(conn, fallbackInstances, lostCapacity)
            ownInstances += fallbackInstances
            otherOwners.update(fallbackOwners)

        return ownInstances, sorted(otherOwners)

    def releaseOwnership(self, machines):
        if machines:
            conn = self.makeEc2Connection()
            instanceIds = [machine.id for machine in machines]
            for inst in conn.get_only_instances(instance_ids=instanceIds):
                inst.remove_tag(self.userTagName)
            self.enableAlarmActions(self.getAlarmNames(instanceIds))

    def getCapacity(self):
        return self.capacity

    def slavesOnRemoteSystem(self):
        return True

    @classmethod
    def findSetUpDirectory(cls, dir):
        # Egg-link points at the Python package code, which may not be all of the checkout
        # Assume the setup.py is where it all starts
        while not os.path.isfile(os.path.join(dir, "setup.py")):
            newDir = os.path.dirname(dir)
            if newDir == dir:
                return
            else:
                dir = newDir
        return dir

    @classmethod
    def findVirtualEnvLinkedDirectories(cls, checkout):
        # "Egg-links" are something found in Python virtual environments
        # They are a sort of portable symbolic link, but of course tools like rsync don't understand them
        # Virtual environments can also point out another environment they were created from, which we may also need to copy
        linkedDirs = []
        realPythonPrefix = sys.real_prefix if hasattr(sys, "real_prefix") else sys.prefix
        for root, _, files in os.walk(checkout):
            for f in sorted(files):
                if f.endswith(".egg-link"):
                    path = os.path.join(root, f)
                    newDir = open(path).read().splitlines()[0].strip()
                    setupDir = cls.findSetUpDirectory(newDir)
                    if setupDir and setupDir not in linkedDirs:
                        linkedDirs.append(setupDir)
                elif f == "orig-prefix.txt":
                    path = os.path.join(root, f)
                    newDir = open(path).read().strip()
                    # Don't try to synch the system Python!
                    if newDir != realPythonPrefix and newDir not in linkedDirs:
                        linkedDirs.append(newDir)
        return linkedDirs

    def getDirectoriesForSynch(self):
        appDir = self.app.getDirectory()
        dirs = [appDir]
        if self.synchSlaveCode():
            dirs.append(plugins.installationRoots[0])
            personalLog = os.getenv("TEXTTEST_PERSONAL_LOG")
            if personalLog:
                dirs.append(personalLog)
        checkout = self.app.getCheckoutForDisplay()
        if checkout and not checkout.startswith(appDir):
            dirs.append(checkout)
            dirs += self.findVirtualEnvLinkedDirectories(checkout)
        return dirs

    def getMachine(self, jobId, includeReleased=False):
        machines = self.machines
        if includeReleased:
            machines = self.machines + self.releasedMachines
        for machine in machines:
            if machine.hasJob(jobId):
                return machine

    def setRemoteProcessId(self, jobId, remotePid):
        machine = self.getMachine(jobId)
        if machine:
            machine.setRemoteProcessId(jobId, remotePid)

    def getRemoteTestMachine(self, jobId):
        machine = self.getMachine(jobId)
        if machine:
            return machine.fullMachine

    def killRemoteProcess(self, jobId):
        machine = self.getMachine(jobId)
        if machine:
            return machine.killRemoteProcess(jobId, self.getSignal())
        else:
            return False, None

    def getJobFailureInfo(self, jobId):
        machine = self.getMachine(jobId, includeReleased=True)
        return machine.errorMessage if machine else ""

    def getStatusForAllJobs(self):
        procStatus = super(QueueSystem, self).getStatusForAllJobs()
        jobStatus = {}
        for machine in self.machines:
            machine.collectJobStatus(jobStatus, procStatus)
        self.cleanup()  # Try to release any machines we're not using
        return jobStatus

    def killJob(self, jobId):
        # ssh doesn't forward signals to remote processes.
        # We need to find it ourselves and send it explicitly. Can assume python exists remotely, but not much else.
        killed, localPid = self.killRemoteProcess(jobId)
        # Hack for self-tests. Shouldn't normally be needed. Need to kill the process locally as well when replaying CaptureMock.
        # Also kill the local process if we can't find the remote one for some reason...
        if localPid and (not killed or os.getenv("CAPTUREMOCK_MODE") == "0"):
            return super(QueueSystem, self).killJob(localPid)
        else:
            return True

    def synchSlaveCode(self):
        # If we're running our self-diagnostics on the slaves, make sure we copy our local code across and run it as the slaves
        return "xs" in self.app.inputOptions

    def getTextTestArgs(self):
        if self.synchSlaveCode():
            return super(QueueSystem, self).getTextTestArgs()
        else:
            return ["texttest"]  # Assume remote nodes are UNIX-based with TextTest installed centrally

    def submitSlaveJob(self, cmdArgs, *args):
        if self.nextMachineIndex >= len(self.machines):
            return None, "No more available machines to submit EC2 jobs to - existing jobs have failed"

        machine = self.machines[self.nextMachineIndex]
        submitter = super(QueueSystem, self).submitSlaveJob
        jobId = machine.submitSlave(submitter, cmdArgs, *args)
        if jobId is None:
            self.releaseOwnership([machine])
            self.machines.remove(machine)
            self.capacity = self.calculateCapacity()
            return self.submitSlaveJob(cmdArgs, *args)
        if machine.isFull():
            self.nextMachineIndex += 1
        return jobId, None


from .local import MachineInfo, getUserSignalKillInfo, getExecutionMachines




import os
import filecmp
import time
import subprocess
import logging
import re
from texttestlib import plugins
from shutil import copyfile

from fnmatch import fnmatch


class FileComparison:
    SAME = 0
    DIFFERENT = 1
    APPROVED = 2

    def __init__(self, test, stem, standardFile, tmpFile, testInProgress=False, **kw):
        self.stdFile = standardFile
        self.stdCmpFile = self.stdFile
        self.tmpFile = tmpFile
        self.tmpCmpFile = tmpFile
        self.stem = stem
        self.differenceCache = self.SAME
        self.recalculationTime = None
        self.diag = logging.getLogger("FileComparison")
        stemForConfig = self.stemForConfig()
        self.severity = test.getCompositeConfigValue("failure_severity", stemForConfig)
        self.displayPriority = test.getCompositeConfigValue("failure_display_priority", stemForConfig)
        maxLength = test.getConfigValue("lines_of_text_difference")
        maxWidth = test.getConfigValue("max_width_text_difference")
        # It would be nice if this could be replaced by some automagic file type detection
        # mechanism, such as the *nix 'file' command, but as the first implementation I've
        # chosen to use a manually created list instead.
        self.binaryFile = test.configValueMatches("binary_file", stemForConfig)
        self.previewGenerator = plugins.PreviewGenerator(maxWidth, maxLength)
        self.textDiffTool = test.getConfigValue("text_diff_program")
        self.textDiffToolMaxSize = plugins.parseBytes(test.getCompositeConfigValue("max_file_size", self.textDiffTool))
        self.freeTextBody = None
        # subclasses may override if they don't want to store in this way
        self.cacheDifferences(test, testInProgress)
        self.diag.info("Created file comparison std: " + repr(self.stdFile) + " tmp: " +
                       repr(self.tmpFile) + " diff: " + repr(self.differenceCache))

    def stemForConfig(self):
        return self.stem

    def setStandardFile(self, standardFile):
        self.stdFile = standardFile
        self.stdCmpFile = self.stdFile
        self.diag.info("Setting standard file for " + self.stem + " to " + repr(standardFile))

    def recompute(self, test):
        self.freeTextBody = None
        if self.needsRecalculation():
            self.recalculationTime = time.time()
        if self.tmpFile:
            if os.path.isfile(self.tmpFile):
                self.cacheDifferences(test, False)
            elif self.differenceCache == self.DIFFERENT:
                # File has been removed
                self.tmpFile = None
                self.tmpCmpFile = None
                self.differenceCache = self.SAME

    def split(self, test, separators):
        separator = separators.get(self.stem)
        if not separator:  # try wildcards
            for key, value in list(separators.items()):
                if fnmatch(self.stem, key):
                    separator = value
                    break
        if separator:
            sepRegex = re.compile(separator)
            tmpParts = self.splitFile(test, self.tmpCmpFile, sepRegex)
            origParts = self.splitFile(test, self.stdCmpFile, sepRegex)
            return [SplitFileComparison(self, test, self.stem, origPart, tmpPart)
                    for origPart, tmpPart in zip(origParts, tmpParts)]
        else:
            return []

    def makeDeltaName(self, name):
        parts = name.rsplit("_", 1)
        lastpart = parts[-1]
        if len(parts) == 2 and lastpart.isdigit():
            newdigit = str(int(lastpart) + 1)
            return parts[0] + "_" + newdigit
        else:
            return name + "_2"

    def getSplitFileName(self, match, partsDir, splitFileNames):
        localname = match.group(1).replace(" ", "_").lower()
        while localname in splitFileNames:
            localname = self.makeDeltaName(localname)
        splitFileNames.append(localname)
        return os.path.join(partsDir, localname)

    def splitFile(self, test, filename, sepRegex):
        parts = []
        localPartsDir = os.path.basename(filename + "_split")
        partsDir = test.makeTmpFileName(localPartsDir, forFramework=1)
        plugins.ensureDirectoryExists(partsDir)
        initialFileName = os.path.join(partsDir, "initial")
        parts.append(initialFileName)
        currWriteFile = open(initialFileName, "w")
        splitFileNames = []
        with open(filename) as f:
            for line in f:
                match = sepRegex.search(line)
                if match:
                    currWriteFile.close()
                    newFileName = self.getSplitFileName(match, partsDir, splitFileNames)
                    parts.append(newFileName)
                    currWriteFile = open(newFileName, "w")
                currWriteFile.write(line)
        currWriteFile.close()
        return parts

    def unsplit(self, test):
        path = test.getDirectory(temporary=True, forFramework=True)
        for dir in self.getSplitDirs(path):
            plugins.rmtree(os.path.join(path, dir))

    def getSplitDirs(self, path):
        dirs = []
        for f in os.listdir(path):
            if os.path.isdir(os.path.join(path, f)) and f.endswith("_split"):
                dirs.append(f)
        return dirs

    def __getstate__(self):
        # don't pickle the diagnostics
        state = {}
        for var, value in list(self.__dict__.items()):
            if var != "diag" and var != "recalculationTime":
                state[var] = value
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self.diag = logging.getLogger("TestComparison")
        self.recalculationTime = None

    def __repr__(self):
        return self.stem

    def modifiedDates(self):
        files = [self.stdFile, self.tmpFile, self.stdCmpFile, self.tmpCmpFile]
        return " : ".join(map(self.modifiedDate, files))

    def modifiedDate(self, file):
        if not file:
            return "---"
        modTime = plugins.modifiedTime(file)
        if modTime:
            return time.strftime("%d%b%H:%M:%S", time.localtime(modTime))
        else:
            return "---"

    def needsRecalculation(self):
        if not self.stdFile or not self.tmpFile:
            self.diag.info("No comparison, no recalculation")
            return False

        # A test that has been approved doesn't need recalculating
        if self.differenceCache == self.APPROVED:
            self.diag.info("Approved file, no recalculation")
            return False

        stdModTime = plugins.modifiedTime(self.stdFile)
        if self.recalculationTime:
            self.diag.info("Already recalculated, checking if file updated since then : " + self.stdFile)
            # If we're already recalculated, only do it again if standard file changes since then
            return stdModTime > self.recalculationTime

        tmpModTime = plugins.modifiedTime(self.tmpFile)
        if stdModTime is not None and tmpModTime is not None and stdModTime >= tmpModTime:
            self.diag.info("Standard result newer than generated result at " + self.stdFile)
            return True

        if self.stdFile == self.stdCmpFile:  # no filters
            return False

        stdCmpModTime = plugins.modifiedTime(self.stdCmpFile)
        self.diag.info("Comparing timestamps for standard files")
        return stdCmpModTime is not None and stdModTime is not None and stdModTime >= stdCmpModTime

    def getType(self):
        return "failure"

    def getDisplayFileName(self):
        if self.newResult():
            return self.tmpFile
        else:
            return self.stdFile

    def getDetails(self):
        # Nothing to report above what is already known
        return ""

    def newResult(self):
        return not self.stdFile and self.tmpFile

    def missingResult(self):
        return self.stdFile and not self.tmpFile

    def isDefunct(self):
        return not self.stdFile and not self.tmpFile

    def hasSucceeded(self):
        return self.stdFile and self.tmpFile and not self.hasDifferences()

    def hasDifferences(self):
        return self.differenceCache == self.DIFFERENT

    def getStdFile(self, filtered, postfix=""):
        if filtered:
            return self.stdCmpFile + postfix
        else:
            return self.stdFile + postfix

    def getTmpFile(self, filtered, postfix=""):
        if filtered:
            return self.tmpCmpFile + postfix
        else:
            return self.tmpFile + postfix

    def existingFile(self, *args):
        if self.missingResult() or self.differenceCache == self.APPROVED:
            return self.getStdFile(*args)
        else:
            return self.getTmpFile(*args)

    def setCmpFiles(self, test, testInProgress):
        filterFileBase = test.makeTmpFileName(self.stem + "." + test.app.name, forFramework=1)
        origCmp = filterFileBase + "origcmp"
        if os.path.isfile(origCmp):
            self.stdCmpFile = origCmp
        tmpCmpFileName = filterFileBase + "cmp"
        if testInProgress:
            tmpCmpFileName = filterFileBase + "partcmp"
        if os.path.isfile(tmpCmpFileName):
            self.tmpCmpFile = tmpCmpFileName

    def updateDifferenceCache(self, valueForEqual):
        if self.stdCmpFile and self.tmpCmpFile:
            if filecmp.cmp(self.stdCmpFile, self.tmpCmpFile, 0):
                if self.differenceCache != self.APPROVED:
                    self.differenceCache = valueForEqual
            else:
                self.differenceCache = self.DIFFERENT
            self.diag.info("Caching differences " + repr(self.stdCmpFile) + " " +
                           repr(self.tmpCmpFile) + " = " + repr(self.differenceCache))

    def cacheDifferences(self, test, testInProgress):
        self.setCmpFiles(test, testInProgress)
        self.updateDifferenceCache(self.SAME)

    def getSummary(self, includeNumbers=True):
        if self.newResult():
            return repr(self) + " new"
        elif self.missingResult():
            return repr(self) + " missing"
        else:
            return self.getDifferencesSummary(includeNumbers)

    def getDifferencesSummary(self, includeNumbers=True):
        return repr(self) + " different"

    def getFreeText(self):
        return self.getFreeTextTitle() + "\n" + self.getFreeTextBody()

    def getFreeTextTitle(self):
        if self.missingResult():
            titleText = "Missing result in"
        elif self.newResult():
            titleText = "New result in"
        else:
            titleText = "Differences in"
        titleText += " " + repr(self)
        return "-" * 10 + " " + titleText + " " + "-" * 10

    def getFreeTextBody(self):
        if self.freeTextBody is None:
            self.freeTextBody = self._getFreeTextBody()
        return self.freeTextBody

    def _getFreeTextBody(self):
        if self.binaryFile and \
                (self.newResult() or self.missingResult()):
            message = "Binary file, not showing any preview. " + \
                      "Edit the configuration entry 'binary_file' and re-run if you suspect that this file contains only text.\n"
            return self.previewGenerator.getWrappedLine(message)
        elif self.newResult():
            return self.previewGenerator.getPreview(open(self.tmpCmpFile, errors="ignore"))
        elif self.missingResult():
            return self.previewGenerator.getPreview(open(self.stdCmpFile, errors="ignore"))

        try:
            stdFileSize = os.path.getsize(self.stdCmpFile)
            tmpFileSize = os.path.getsize(self.tmpCmpFile)
            if self.textDiffToolMaxSize >= 0 and \
                (stdFileSize > self.textDiffToolMaxSize or
                    tmpFileSize > self.textDiffToolMaxSize):
                message = "The result files were too large to compare - " + str(stdFileSize) + " and " + \
                          str(tmpFileSize) + " bytes, compared to the limit of " + str(self.textDiffToolMaxSize) + \
                          " bytes. Adjust the configuration entry 'max_file_size' for the tool '" + self.textDiffTool + \
                          "' and re-run to see the difference in this text view.\n"
                return self.previewGenerator.getWrappedLine(message)

            cmdArgs = plugins.splitcmd(self.textDiffTool) + [self.stdCmpFile, self.tmpCmpFile]
            proc = subprocess.Popen(cmdArgs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
            return self.previewGenerator.getPreview(proc.stdout)
        except OSError as e:
            self.diag.info("No diff report: full exception printout\n" + plugins.getExceptionString())
            return "No difference report could be created: could not find textual difference tool '" + self.textDiffTool + "'\n" + \
                   "(" + str(e) + ")"

    def updateAfterLoad(self, changedPaths):
        for oldPath, newPath in changedPaths:
            if self.stdFile:
                self.stdFile = self.stdFile.replace(oldPath, newPath)
                self.stdCmpFile = self.stdCmpFile.replace(oldPath, newPath)
            if self.tmpFile:
                self.tmpCmpFile = self.tmpCmpFile.replace(oldPath, newPath)
                self.tmpFile = self.tmpFile.replace(oldPath, newPath)

    def versionise(self, fileName, versionString):
        if versionString:
            return fileName + "." + versionString
        else:
            return fileName

    def getStdRootVersionFile(self):
        # drop version identifiers
        dirname, local = os.path.split(self.stdFile)
        localRoot = ".".join(local.split(".")[:2])
        return os.path.join(dirname, localRoot)

    def versionMatchesStd(self, versionString):
        if versionString is None:
            return True
        versions = set(versionString.split("."))
        local = os.path.basename(self.stdFile)
        localVersions = set(local.split(".")[2:])
        return versions == localVersions

    def getStdFileForSave(self, versionString):
        if self.versionMatchesStd(versionString):
            return self.stdFile
        else:
            stdRoot = self.getStdRootVersionFile()
            return self.versionise(stdRoot, versionString)

    def backupOrRemove(self, fileName, backupVersionStrings):
        if os.path.isfile(fileName):
            for backupVersionString in backupVersionStrings:
                backupFile = self.getStdFileForSave(backupVersionString)
                if not os.path.isfile(backupFile):
                    copyfile(fileName, backupFile)
            os.remove(fileName)

    def overwrite(self, test, exact, versionString, backupVersionStrings):
        self.diag.info("save file from " + self.tmpFile)
        self.stdFile = self.getStdFileForSave(versionString)
        self.backupOrRemove(self.stdFile, backupVersionStrings)
        self.saveTmpFile(test, exact)

    def overwriteFromSplit(self, splitComps, test, exact, versionString, backupVersionStrings):
        self.stdFile = self.getStdFileForSave(versionString)
        self.diag.info("writing split files back to " + self.stdFile)
        self.freeTextBody = None  # clear the cache which may well be wrong now...
        self.backupOrRemove(self.stdFile, backupVersionStrings)
        with open(self.stdFile, "w") as f:
            for splitComp in splitComps:
                f.write(open(splitComp.stdFile).read())
        self.stdCmpFile = self.stdFile
        self.updateDifferenceCache(self.APPROVED)

    def saveNew(self, test, versionString):
        self.stdFile = os.path.join(test.getDirectory(), self.versionise(
            self.stem + "." + test.app.name, versionString))
        self.saveTmpFile(test)

    def getTmpFileForSave(self, test):
        if not test.configValueMatches("save_filtered_file_stems", self.stemForConfig()):
            return self.tmpFile

        # Don't include the sorting when saving filtered files...
        normalFile = self.tmpCmpFile + ".normal"
        if os.path.isfile(normalFile):
            return normalFile
        else:
            return self.tmpCmpFile

    def saveTmpFile(self, test, exact=True):
        self.diag.info("Saving tmp file to " + self.stdFile + ", exact=" + repr(exact))
        plugins.ensureDirExistsForFile(self.stdFile)
        # Allow for subclasses to differentiate between a literal overwrite and a
        # more intelligent save, e.g. for performance. Default is the same for exact
        # and inexact save
        tmpFile = self.getTmpFileForSave(test)
        if os.path.isfile(tmpFile):
            if exact:
                copyfile(tmpFile, self.stdFile)
            else:
                self.saveResults(tmpFile, self.stdFile)
        else:
            self.diag.info("Failed to save, no file at " + tmpFile)
            raise plugins.TextTestError(
                "The following file seems to have been removed since it was created:\n" + repr(tmpFile))
        self.differenceCache = self.APPROVED

    def saveMissing(self, versionString, autoGenText, backupVersionStrings):
        stdRoot = self.getStdRootVersionFile()
        self.diag.info("Saving missing file for " + stdRoot + " with version " + repr(versionString))
        targetFile = self.versionise(stdRoot, versionString)
        self.backupOrRemove(targetFile, backupVersionStrings)

        if self.stdFile != targetFile and os.path.isfile(self.stdFile):
            # Create a "versioned-missing" file
            newFile = open(targetFile, "wb")
            newFile.write(autoGenText)
            newFile.close()
        self.stdFile = None
        self.stdCmpFile = None

    def saveResults(self, tmpFile, destFile):
        copyfile(tmpFile, destFile)

    def getParent(self):
        pass


class SplitFileComparison(FileComparison):
    def __init__(self, parent, test, stem, stdFile, *args):
        self.parent = parent
        stemToUse = stem + "/" + os.path.basename(stdFile)
        FileComparison.__init__(self, test, stemToUse, stdFile, *args)

    def getParent(self):
        return self.parent

    def setCmpFiles(self, *args):
        pass  # Don't want to look for comparison files

    def needsRecalculation(self):
        return False  # These cannot be recalculated in any sensible way, and are created nearly simultaneously




""" Base class for all the queue system implementations """

import subprocess
import os
import signal
from . import abstractqueuesystem
from multiprocessing import cpu_count
from texttestlib import plugins


class QueueSystem(abstractqueuesystem.QueueSystem):
    def __init__(self, *args):
        self.processes = {}

    def submitSlaveJob(self, cmdArgs, slaveEnv, logDir, submissionRules, jobType):
        outputFile, errorsFile = submissionRules.getJobFiles()
        stdout = open(os.path.join(logDir, outputFile), "w")
        stderr = open(os.path.join(logDir, errorsFile), "w")
        createflags = subprocess.CREATE_NEW_PROCESS_GROUP if os.name == "nt" else 0
        try:
            process = subprocess.Popen(cmdArgs, stdout=stdout, stderr=stderr,
                                       cwd=logDir, env=self.getSlaveEnvironment(slaveEnv),
                                       startupinfo=plugins.getHideStartUpInfo(),
                                       creationflags=createflags)
            errorMessage = None
        except OSError as e:
            stdout.close()
            stderr.close()
            errorMessage = "Failed to start slave process : " + str(e)
        if errorMessage:
            return None, self.getFullSubmitError(errorMessage, cmdArgs, jobType)
        else:
            jobId = str(process.pid)
            self.processes[jobId] = process
            return jobId, None

    def getCapacity(self):
        return cpu_count()

    def formatCommand(self, cmdArgs):
        return " ".join(cmdArgs)

    def getSignal(self):
        return signal.SIGUSR2 if os.name == "posix" else signal.CTRL_C_EVENT

    def killJob(self, jobId):
        proc = self.processes[jobId]
        jobExisted = proc.poll() is None
        if jobExisted:
            proc.send_signal(self.getSignal())
        return jobExisted

    def runTaskKill(self, proc, extraArgs=[]):
        return subprocess.call(["taskkill"] + extraArgs + ["/PID", str(proc.pid)], stdout=open(os.devnull, "w"), stderr=subprocess.STDOUT, startupinfo=plugins.getHideStartUpInfo()) == 0

    def getStatusForAllJobs(self):
        statusDict = {}
        for procId, process in list(self.processes.items()):
            if process.poll() is None:
                statusDict[procId] = "RUN", "Running"
        return statusDict

    def getJobFailureInfo(self, jobId):
        return ""  # no accounting system here...

    def getQueueSystemName(self):
        return "local queue"

# Interpret what the limit signals mean...


def getUserSignalKillInfo(userSignalNumber, explicitKillMethod):
    return explicitKillMethod()

# Used by slave for producing performance data


class MachineInfo:
    def findActualMachines(self, machineOrGroup):
        # In LSF this unpacks host groups, taking advantage of the fact that they are
        # interchangeable with machines. This is not true in SGE anyway, so don't support it.
        return [machineOrGroup]

    def findResourceMachines(self, resource):
        return []

    def findRunningJobs(self, machine):
        return []


def getExecutionMachines():
    from texttestlib.plugins import gethostname
    return [gethostname()]




# This file (at the moment) has the sole purpose of specifying the texttest version number.
# It's in a separate file to make it easy to find.

version = "4.0.8"

# Note: Decided it's not a good idea to require debug versions here. Even if we have Python 2.4.3 it's
# pretty likely 2.4.2 works OK unless we know otherwise. The interface shouldn't be different at least.

# Which python version do we require?
required_python_version = (3, 6, 0)

# Which pygtk version do we require?
required_pygtk_version = (3, 18, 0)



import datetime
import time
import os
from texttestlib import plugins
from . import vcs_independent


DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'


class GitInterface(vcs_independent.VersionControlInterface):
    def __init__(self, controlDir):
        self.warningStateInfo = {"M": "Modified", "D": "Deleted", "A": "Added", "R": "Renamed"}
        self.errorStateInfo = {"??": "Unknown"}
        # "!!" for ignored files is only available on git versions < 1.7.4
        self.allStateInfo = {"C": "Copied", "U": "Unmerged"}
        self.allStateInfo.update(self.warningStateInfo)
        self.allStateInfo.update(self.errorStateInfo)
        self.vcsDirectory = os.path.dirname(controlDir)
        vcs_independent.VersionControlInterface.__init__(self, controlDir, "Git",
                                                         list(self.warningStateInfo.values()), list(self.errorStateInfo.values()), "HEAD")
        self.defaultArgs["rm"] = ["--force", "-r"]
        # Would like to use --ignored but it is not available on git versions < 1.7.4
        self.defaultArgs["status"] = ["--porcelain"]
        self.defaultArgs["log"] = ["-p", "--follow", "--date=format:%s" % DATE_FORMAT]

    def getDateFromLog(self, output):
        for line in output.splitlines():
            if line.startswith("Date:"):
                return datetime.datetime.strptime(line.split()[1], DATE_FORMAT)

    def getGraphicalDiffArgs(self, diffProgram):
        return ["git", "difftool", "-t", diffProgram, "-y"]

    def startGUIProcess(self, cmdArgs, **kw):
        relArgs = list(map(self.makeRelPath, cmdArgs))
        vcs_independent.VersionControlInterface.startGUIProcess(self, relArgs, **kw)

    def getStateFromStatus(self, output):
        words = output.split()
        if len(words) > 0:
            statusLetter = words[0]
            return self.allStateInfo.get(statusLetter, statusLetter)
        else:
            return "Unchanged"

    def getFileNames(self, fileArg, recursive, forStatus=False, **kwargs):
        # Git handles ignored files different. We have to remove all ignored files to avoid doing status on them
        fileNames = vcs_independent.VersionControlInterface.getFileNames(self, fileArg, recursive, **kwargs)
        if not forStatus:
            return fileNames

        ignored = self.getIgnoredFiles(fileArg)
        return [f for f in fileNames if self.makeRelPath(f) not in ignored]

    def getIgnoredFiles(self, path):
        if not os.path.isdir(path):
            return []
        _, stdout, _ = self.getProcessResults(["git", "ls-files", "--other", "-i", "--exclude-standard",  path])
        return stdout.split()

    def makeRelPath(self, arg):
        if os.path.isabs(arg):
            relpath = plugins.relpath(arg, self.vcsDirectory)
            if relpath:
                return relpath
        return arg

    def getProcessResults(self, args, cwd=None, **kwargs):
        workingDir = cwd if cwd else self.vcsDirectory
        if workingDir == self.vcsDirectory:
            args = list(map(self.makeRelPath, args))
        return vcs_independent.VersionControlInterface.getProcessResults(self, args, cwd=workingDir, **kwargs)

    def callProgram(self, cmdName, fileArgs=[], **kwargs):
        if fileArgs:
            fileArgs = list(map(os.path.realpath, fileArgs))
        return vcs_independent.VersionControlInterface.callProgram(self, cmdName, fileArgs, cwd=self.vcsDirectory, **kwargs)

    def getCombinedRevisionOptions(self, r1, r2):
        return [r1 + ".." + r2, "--"]

    def getSingleRevisionOptions(self, r1):
        return [r1]

    def removePath(self, path):
        # Git doesn't remove unknown files
        retCode = self.callProgram("rm", [path])
        plugins.removePath(path)
        return retCode == 0

    def hasLocalCommits(self, vcsDirectory):
        retCode, _, stderr = self.getProcessResults(["git", "push", "-n"], cwd=vcsDirectory)
        return retCode == 0 and stderr.strip() != "Everything up-to-date"


vcs_independent.vcsClass = GitInterface


class FilteredDiffGUI(vcs_independent.FilteredDiffGUI):
    def __init__(self, *args):
        vcs_independent.FilteredDiffGUI.__init__(self, *args)
        self.cmdName = "show"

    def getTmpFileArgs(self, fileName, revision):
        revStr = revision or "master"
        return [revStr + ":" + vcs_independent.vcs.makeRelPath(fileName)]


class FilteredDiffGUIRecursive(FilteredDiffGUI):
    recursive = True


class UpdateGUI(vcs_independent.UpdateGUI):
    def getCommandName(self):
        return "pull"

    @staticmethod
    def _getTitle():
        return "Pull"


class InteractiveActionConfig(vcs_independent.InteractiveActionConfig):
    def diffClasses(self):
        return [vcs_independent.DiffGUI, vcs_independent.DiffGUIRecursive, FilteredDiffGUI, FilteredDiffGUIRecursive]

    def getUpdateClass(self):
        return UpdateGUI




""" 
The various "holders" for displaying the "ActionGUI" abstraction: i.e. menus, toolbars and popups,
to store the simple actions and the dialogs, and a notebook to store the tabs in.
"""
from gi.repository import Gtk
import os
import sys
import logging
import types
from . import guiutils
from texttestlib import plugins
from collections import OrderedDict
from pprint import pformat
from functools import cmp_to_key


class MenuBarGUI(guiutils.SubGUI):
    def __init__(self, dynamic, uiManager, actionGUIs, menuNames, *args):
        guiutils.SubGUI.__init__(self)
        # Create GUI manager, and a few default action groups
        self.menuNames = menuNames
        self.dynamic = dynamic
        self.allFiles = plugins.findDataPaths(["*.xml"], *args)
        self.uiManager = uiManager
        self.actionGUIs = actionGUIs
        self.actionGroup = self.uiManager.get_action_groups()[0]
        self.toggleActions = []
        self.loadedModules = set()
        self.diag = logging.getLogger("Menu Bar")
        self.diag.info("All description files : " + pformat(self.allFiles))

    def getLoadedModules(self):
        if not self.loadedModules:
            self.loadedModules = set((module.split(".")[-1] for module in list(sys.modules.keys())))
        return self.loadedModules

    def shouldHide(self, name):
        return guiutils.guiConfig.getCompositeValue("hide_gui_element", name, modeDependent=True)

    def toggleVisibility(self, action, observer, *args):
        widget = observer.widget
        oldVisible = widget.get_property('visible')
        newVisible = action.get_active()
        if oldVisible and not newVisible:
            widget.hide()
        elif newVisible and not oldVisible:
            widget.show()

    def createToggleActions(self):
        for observer in self.observers:
            actionTitle = observer.getWidgetName()
            actionName = actionTitle.replace("_", "")
            gtkAction = Gtk.ToggleAction(actionName, actionTitle, None, None)
            if observer.shouldShow():
                gtkAction.set_active(True)
            else:
                gtkAction.set_sensitive(False)
            self.actionGroup.add_action(gtkAction)
            gtkAction.connect("toggled", self.toggleVisibility, observer)
            self.toggleActions.append(gtkAction)

    def createView(self):
        # Initialize
        for menuName in self.menuNames:
            realMenuName = menuName
            if not menuName.isupper():
                realMenuName = menuName.capitalize()
            self.actionGroup.add_action(Gtk.Action(menuName + "menu", "_" + realMenuName, None, None))
        self.createToggleActions()

        for file in self.getGUIDescriptionFileNames():
            try:
                self.diag.info("Reading UI from file " + file)
                self.uiManager.add_ui_from_file(file)
            except Exception as e:
                raise plugins.TextTestError("Failed to parse GUI description file '" + file + "': " + str(e))
        self.uiManager.ensure_update()
        self.widget = self.uiManager.get_widget("/MainMenuBar")
        return self.widget

    def notifyTopWindow(self, window):
        window.add_accel_group(self.uiManager.get_accel_group())
        if self.shouldHide("menubar"):
            self.widget.hide()
        for toggleAction in self.toggleActions:
            if self.shouldHide(toggleAction.get_name()):
                toggleAction.set_active(False)

    def getGUIDescriptionFileNames(self):
        # Pick up all GUI descriptions corresponding to modules we've loaded
        loadFiles = list(filter(self.shouldLoad, self.allFiles))
        loadFiles.sort(key=cmp_to_key(self.cmpDescFiles))
        return loadFiles

    def cmpDescFiles(self, file1, file2):
        basicFiles = ["default_gui.xml", "default_gui-dynamic.xml", "default_gui-static.xml"]
        base1 = os.path.basename(file1)
        base2 = os.path.basename(file2)
        default1 = base1 in basicFiles
        default2 = base2 in basicFiles
        if default1 != default2:
            # Hard code the three files above, they define the basic framework and should come first
            return (default2 > default1) - (default2 < default1)
        partCount1 = base1.count("-")
        partCount2 = base2.count("-")
        if partCount1 != partCount2:
            # less - implies read first (not mode-specific)
            return (partCount1 > partCount2) - (partCount1 < partCount2)
        # something deterministic, just to make sure it's the same for everyone
        return (base2 > base1) - (base2 < base1)

    def shouldLoad(self, fileName):
        # Infer whether a file should be loaded
        # The format is
        # <module_name>-<config setting>-<config setting...>-[|dynamic|static].xml
        allParts = os.path.basename(fileName)[:-4].split("-")  # drop .xml, split on "-"
        moduleName = allParts[0]
        if allParts[-1] == "dynamic":
            if self.dynamic:
                allParts = allParts[:-1]
            else:
                return False
        elif allParts[-1] == "static":
            if self.dynamic:
                return False
            else:
                allParts = allParts[:-1]

        self.diag.info("Checking if we loaded module " + moduleName)
        if moduleName not in self.getLoadedModules():
            return False

        for configSetting in allParts[1:]:
            value = guiutils.guiConfig.getValue(configSetting)
            haveSet = self.haveSet(value)
            self.diag.info("Checking if we have set " + configSetting +
                           ": value = " + repr(value) + ", set = " + repr(haveSet))
            if not haveSet:
                return False
        return True

    def haveSet(self, val):
        if type(val) == dict:
            if len(val) == 1:
                if "" in val:
                    return val[""]
                elif "default" in val:
                    return False  # Not set if we only have a default key...
            else:
                return True
        else:
            return val and val != "disabled"


class ToolBarGUI(guiutils.ContainerGUI):
    def __init__(self, uiManager, subgui):
        guiutils.ContainerGUI.__init__(self, [subgui])
        self.uiManager = uiManager

    def getWidgetName(self):
        return "_Toolbar"

    def ensureVisible(self, toolbar):
        for item in toolbar.get_children():
            # Or newly added children without stock ids won't be visible in Gtk.ToolbarStyle.BOTH_HORIZ style
            item.set_is_important(True)

    def shouldShow(self):
        return True  # don't care about whether we have a progress bar or not

    def createView(self):
        self.uiManager.ensure_update()
        toolbar = self.uiManager.get_widget("/MainToolBar")
        self.ensureVisible(toolbar)

        # replaced HandleBox with Box and widget.add with widget.pack_start, needs review MB 2018-12-07
        self.widget = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.widget.pack_start(toolbar, True, True, 0)
        progressBarGUI = self.subguis[0]
        if progressBarGUI.shouldShow():
            progressBar = progressBarGUI.createView()
            width = 7  # Looks good, same as Gtk.Paned border width
            alignment = Gtk.Alignment.new(1.0, 1.0, 1.0, 1.0)
            alignment.set_padding(width, width, 1, width)
            alignment.add(progressBar)
            toolItem = Gtk.ToolItem()
            toolItem.add(alignment)
            toolItem.set_expand(True)
            toolbar.insert(toolItem, -1)

        self.widget.show_all()
        return self.widget


def createPopupGUIs(uiManager):
    return PopupMenuGUI("TestPopupMenu", uiManager), PopupMenuGUI("TestFilePopupMenu", uiManager), PopupMenuGUI("ConfigFilePopupMenu", uiManager)


class PopupMenuGUI(guiutils.SubGUI):
    def __init__(self, name, uiManager):
        guiutils.SubGUI.__init__(self)
        self.name = name
        self.uiManager = uiManager

    def createView(self):
        self.uiManager.ensure_update()
        self.widget = self.uiManager.get_widget("/" + self.name)
        self.widget.show_all()
        return self.widget

    def showMenu(self, treeview, event):
        if event.button == 3 and len(self.widget.get_children()) > 0:
            time = event.time
            pathInfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            selection = treeview.get_selection()
            selectedRows = selection.get_selected_rows()
            # If they didnt right click on a currently selected
            # row, change the selection
            if pathInfo is not None:
                if pathInfo[0] not in selectedRows[1]:
                    selection.unselect_all()
                    selection.select_path(pathInfo[0])
                treeview.grab_focus()
                self.widget.popup(None, None, None, None, event.button, time)
                treeview.emit_stop_by_name("button-press-event")


class NotebookGUI(guiutils.SubGUI):
    def __init__(self, tabInfo):
        guiutils.SubGUI.__init__(self)
        self.diag = logging.getLogger("GUI notebook")
        self.tabInfo = [tabGUI for tabGUI in tabInfo if tabGUI.shouldShow()]
        self.notebook = None
        self.currentTabGUI = self.findInitialCurrentTab()
        self.diag.info("Current page set to '" + self.currentTabGUI.getTabTitle() + "'")

    def createView(self):
        self.notebook = Gtk.Notebook()
        self.notebook.set_name("main right-hand notebook")
        for tabGUI in self.tabInfo:
            tabName = tabGUI.getTabTitle()
            label = Gtk.Label(label=tabName)
            page = self.createPage(tabGUI, tabName)
            self.notebook.append_page(page, label)

        self.notebook.set_scrollable(True)
        self.notebook.show()
        self.notebook.connect("switch-page", self.pageSwitched)
        return self.notebook

    def createPage(self, tabGUI, tabName):
        self.diag.info("Adding page " + tabName)
        page = tabGUI.createView()
        if not tabGUI.shouldShowCurrent():
            self.diag.info("Hiding page " + tabName)
            page.hide()
        return page

    def findInitialCurrentTab(self):
        for tabGUI in self.tabInfo:
            if tabGUI.shouldShowCurrent():
                return tabGUI

    def findFirstRemaining(self, pagesRemoved):
        for page in self.notebook.get_children():
            if page.get_property("visible"):
                pageNum = self.notebook.page_num(page)
                if pageNum not in pagesRemoved:
                    return pageNum

    def showNewPages(self, *args):
        changed = False
        for pageNum, tabGUI in enumerate(self.tabInfo):
            page = self.notebook.get_nth_page(pageNum)
            if page is None:
                continue  # Can happen if tests terminate when the GUI is being taken down
            name = tabGUI.getTabTitle()
            if tabGUI.shouldShowCurrent(*args):
                if not page.get_property("visible"):
                    self.diag.info("Showing page " + name)
                    page.show()
                    changed = True
            else:
                self.diag.info("Remaining hidden " + name)
        return changed

    def pageSwitched(self, dummy, dummy2, newNum, *args):
        newTabGUI = self.tabInfo[newNum]
        self.diag.info("Resetting current page to page " + repr(newNum) + " = " + repr(newTabGUI.getTabTitle()))
        # Must do this afterwards, otherwise the above change doesn't propagate
        self.currentTabGUI = newTabGUI
        self.diag.info("Resetting done.")

    def findPagesToHide(self, *args):
        pages = OrderedDict()
        for pageNum, tabGUI in enumerate(self.tabInfo):
            page = self.notebook.get_nth_page(pageNum)
            if not tabGUI.shouldShowCurrent(*args) and page.get_property("visible"):
                pages[pageNum] = page
        return pages

    def hideOldPages(self, *args):
        # Must reset the current page before removing it if we're viewing a removed page
        # otherwise we can output lots of pages we don't really look at
        pagesToHide = self.findPagesToHide(*args)
        if len(pagesToHide) == 0:
            return False

        if self.notebook.get_current_page() in pagesToHide:
            newCurrentPageNum = self.findFirstRemaining(pagesToHide)
            if newCurrentPageNum is not None:
                self.notebook.set_current_page(newCurrentPageNum)

        # remove from the back, so we don't momentarily view them all if removing everything
        for page in reversed(list(pagesToHide.values())):
            self.diag.info("Hiding page " + self.notebook.get_tab_label_text(page))
            page.hide()
        return True

    def updateCurrentPage(self, rowCount):
        for pageNum, tabGUI in enumerate(self.tabInfo):
            if tabGUI.shouldShowCurrent() and tabGUI.forceVisible(rowCount):
                return self.notebook.set_current_page(pageNum)

    def notifyNewTestSelection(self, tests, dummyApps, rowCount, direct):
        # This is mostly an attempt to work around the tree search problems.
        # Don't hide the tab for user-deselections of all tests because it trashes the search.
        if len(tests) > 0 or not direct:
            self.diag.info("New selection of size " + str(len(tests)) + ", adjusting notebook")
            # only change pages around if a test is directly selected and we haven't already selected another important tab
            changeCurrentPage = direct and not self.currentTabGUI.forceVisible(rowCount)
            self.diag.info("Current tab gui " + repr(self.currentTabGUI.__class__) +
                           " will change = " + repr(changeCurrentPage))
            self.updatePages(rowCount=rowCount, changeCurrentPage=changeCurrentPage)

    def updatePages(self, test=None, state=None, rowCount=0, changeCurrentPage=False):
        if not self.notebook:
            return
        self.showNewPages(test, state)
        self.hideOldPages(test, state)
        if changeCurrentPage:
            self.updateCurrentPage(rowCount)

    def notifyLifecycleChange(self, test, state, *args):
        self.updatePages(test, state)

    def addSuites(self, *args):
        self.updatePages()




import socket
import os

def getPortListenErrorCode(ip, port):
    if "TEXTTEST_FAKE_USER" in os.environ:
        return 0
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(5)
    ret = s.connect_ex((ip, port))
    s.close()
    return ret


def getUserName():
    return os.getenv("TEXTTEST_FAKE_USER", os.getenv("USER", os.getenv("USERNAME"))) # Fake user id, useful for testing cloud interaction



#!/usr/local/bin/python

import os
import sys
import subprocess
import signal
import logging
from texttestlib import plugins

# Unlike earlier incarnations of this functionality,
# we don't rely on sharing displays but create our own for each test run.


class VirtualDisplayResponder(plugins.Responder):
    instance = None

    def __init__(self, *args):
        plugins.Responder.__init__(self, *args)
        self.displayInfoList = []
        self.guiSuites = []
        self.diag = logging.getLogger("virtual display")
        self.killDiag = logging.getLogger("kill processes")
        VirtualDisplayResponder.instance = self

    def addSuites(self, suites):
        guiSuites = [suite for suite in suites if suite.getConfigValue("use_case_record_mode") == "GUI"]
        if len(self.displayInfoList) == 0:
            self.setUpVirtualDisplay(guiSuites)
            for var, value in self.getVariablesToSet():
                plugins.log.info("Tests will run with " + var + " variable set to " + value)

    def getVariablesToSet(self):
        vars = []
        for i, (_, displayName, _, _, _, _) in enumerate(self.displayInfoList):
            suffix = "" if i == 0 else str(i + 1)
            vars.append(("DISPLAY" + suffix, displayName))
        return vars

    def setUpVirtualDisplay(self, guiSuites):
        if len(guiSuites) == 0:
            return
        machines = self.findMachines(guiSuites)
        displayCount = max((suite.getConfigValue("virtual_display_count") for suite in guiSuites))
        for _ in range(displayCount):
            displayInfo = self.getDisplayInfo(machines, guiSuites[0].app)
            if displayInfo:
                self.displayInfoList.append(displayInfo)
                self.guiSuites = guiSuites
            elif len(machines) > 0:
                plugins.printWarning("Failed to start virtual display on " +
                                     ",".join(machines) + " - using real display.")

    def getDisplayInfo(self, machines, app):
        for machine in machines:
            displayName, xvfbPid, xvfbOrSshProc, wmPid, wmOrSshProc = self.createDisplay(machine, app)
            if displayName:
                return machine, displayName, xvfbPid, xvfbOrSshProc, wmPid, wmOrSshProc
            else:
                plugins.printWarning("Virtual display program Xvfb not available on " + machine, stdout=True)

    def findMachines(self, suites):
        allMachines = []
        for suite in suites:
            for machine in suite.getConfigValue("virtual_display_machine"):
                if machine == "localhost":  # Local to the test run, not to where we're trying to run...
                    machine = suite.app.getRunMachine()
                if not machine in allMachines:
                    allMachines.append(machine)
        return allMachines

    def notifyTestProcessComplete(self, test):
        if self.restartXvfbAndWm():
            plugins.log.info("Virtual display had terminated unexpectedly with some test processes still to run.")
            for var, value in self.getVariablesToSet():
                plugins.log.info("Reset " + var + " variable to " + value + " for test " + repr(test))
                test.setEnvironment(var, value)

    def notifyComplete(self, *args):
        if self.restartXvfbAndWm():
            plugins.log.info("Virtual display had terminated unexpectedly.")
            for var, value in self.getVariablesToSet():
                plugins.log.info("Reset " + var + " variable to " + value + ".")

    def restartXvfbAndWm(self):
        # Whenever a test completes, we check to see if the virtual server is still going
        changed = False
        for i, displayInfo in enumerate(self.displayInfoList):
            xvfbOrSshProc = displayInfo[3]
            if xvfbOrSshProc is not None and xvfbOrSshProc.poll() is not None:
                xvfbOrSshProc.wait()  # Don't leave zombie processes around
                # If Xvfb has terminated, we need to restart it
                newDisplayInfo = self.getDisplayInfo(self.findMachines(self.guiSuites), self.guiSuites[0].app)
                if newDisplayInfo:
                    self.displayInfoList[i] = newDisplayInfo
                    changed = True
        return changed

    def notifyAllComplete(self):
        self.cleanXvfbAndWm()

    def terminateIfRunning(self, pid):
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError:  # pragma: no cover - only set up this way to avoid race conditions. Should never happen in real life anyway
            pass

    def cleanXvfbAndWm(self):
        if len(self.displayInfoList) and os.name == "posix":
            wmExecutable = self.guiSuites[0].app.getConfigValue("virtual_display_wm_executable")
            for displayInfo in self.displayInfoList:
                machine, _, xvfbPid, xvfbOrSshProc, wmPid, wmOrSshProc = displayInfo
                if wmExecutable:
                    self.killProcess("window manager", machine, wmPid, wmOrSshProc)
                self.killProcess("Xvfb", machine, xvfbPid, xvfbOrSshProc)
            self.displayInfoList = []

    def killProcess(self, procName, machine, pid, localProc):
        if machine == "localhost":
            self.killDiag.info("Killing " + procName + " process " + str(pid))
            self.terminateIfRunning(pid)
        else:
            self.killDiag.info("Killing remote " + procName + " process on " + machine + " with pid " + str(pid))
            self.guiSuites[0].app.runCommandOn(machine, ["kill", str(pid)])
            # only for self-tests really : traffic mechanism doesn't fake remote process
            self.terminateIfRunning(localProc.pid)
        localProc.wait()  # don't leave zombies around

    def createDisplay(self, machine, app):
        if not self.executableExists(machine, app, "Xvfb"):
            return None, None, None, None, None
        wmExecutable = app.getConfigValue("virtual_display_wm_executable")
        if wmExecutable and not self.executableExists(machine, app, wmExecutable):
            return None, None, None, None, None

        extraArgs = plugins.splitcmd(app.getConfigValue("virtual_display_extra_args"))
        xvfbCmd = self.getRemoteCmd(machine, app, "startXvfb.py", extraArgs)
        displayName, xvfbPid, xvfbOrSshProc = self.startXvfb(xvfbCmd, machine)

        wmPid, wmOrSshProc = None, None
        if wmExecutable:
            extraArgs = [wmExecutable, displayName]
            wmArgs = self.getRemoteCmd(machine, app, "startWindowManager.py", extraArgs)
            wmPid, wmOrSshProc = self.startWindowManager(wmArgs, machine)
        return displayName, xvfbPid, xvfbOrSshProc, wmPid, wmOrSshProc

    def ignoreSignals(self):
        for signum in [signal.SIGUSR1, signal.SIGUSR2, signal.SIGXCPU]:
            signal.signal(signum, signal.SIG_IGN)

    def startXvfb(self, command, machine):
        for _ in range(5):
            self.diag.info("Starting Xvfb using args " + repr(command))
            # Ignore job control signals for remote processes
            # Otherwise the ssh process gets killed, but the stuff it's started remotely doesn't, and we leak Xvfb processes
            preexec_fn = None if machine == "localhost" else self.ignoreSignals
            xvfbOrSshProc = subprocess.Popen(command, preexec_fn=preexec_fn, stdin=open(
                os.devnull), stdout=subprocess.PIPE, stderr=open(os.devnull, "w"), universal_newlines=True)
            line = plugins.retryOnInterrupt(xvfbOrSshProc.stdout.readline)
            if "Time Out!" in line:
                xvfbOrSshProc.wait()
                xvfbOrSshProc.stdout.close()
                self.diag.info("Timed out waiting for Xvfb to come up")
                # We try again and hope for a better process ID!
                continue
            try:
                displayNum, xvfbPid = list(map(int, line.strip().split(",")))
                xvfbOrSshProc.stdout.close()
                return self.getDisplayName(machine, displayNum), xvfbPid, xvfbOrSshProc
            except ValueError:  # pragma : no cover - should never happen, just a fail-safe
                sys.stderr.write("ERROR: Failed to parse startXvfb.py line :\n " + line + "\n")
                xvfbOrSshProc.stdout.close()
                return None, None, None

        messages = "Failed to start Xvfb in 5 attempts, giving up"
        plugins.printWarning(messages)
        return None, None, None

    def getPersonalLogDir(self):
        if not self.diag.isEnabledFor(logging.INFO):
            return os.devnull

        return os.getenv("TEXTTEST_PERSONAL_LOG", os.devnull)

    def getRemoteCmd(self, machine, app, pythonScript, extraArgs):
        binDir = plugins.installationDir("libexec")
        localPath = os.path.join(binDir, pythonScript)
        appTmpDir = app.getRemoteTmpDirectory()[1]
        if appTmpDir:
            logDir = os.path.join(appTmpDir, "Xvfb")
            app.ensureRemoteDirExists(machine, logDir)
            remotePath = os.path.join(appTmpDir, pythonScript)
            app.copyFileRemotely(localPath, "localhost", remotePath, machine)
            fullPath = remotePath
            pythonArgs = ["python", "-u"]
        else:
            logDir = self.getPersonalLogDir()
            fullPath = localPath
            pythonArgs = self.findPythonArgs(machine)

        cmdArgs = pythonArgs + [fullPath, logDir] + extraArgs
        return app.getCommandArgsOn(machine, cmdArgs)

    def findPythonArgs(self, machine):
        # In case it isn't the default, allow for a ttpython script in the installation
        if machine == "localhost":
            return [sys.executable, "-u"]

        localPointer = plugins.installationPath("bin/ttpython")
        if localPointer:
            return [localPointer, "-u"]
        else:  # pragma : no cover -there is one in our local installation whether we like it or not...
            return ["python", "-u"]

    def getDisplayName(self, machine, displayNumber):
        # No point in using the port if we don't have to, this seems less reliable if the process is local
        # X keeps track of these numbers internally and connecting to them works rather better.
        displayStr = ":" + str(displayNumber) + ".0"
        if machine == "localhost":
            return displayStr
        else:
            # Don't include user name, if any
            return machine.split("@")[-1] + displayStr

    def startWindowManager(self, command, machine):
        self.diag.info("Starting window manager")
        preexec_fn = None if machine == "localhost" else self.ignoreSignals
        wmProc = subprocess.Popen(command, preexec_fn=preexec_fn, stdin=open(
            os.devnull), stdout=subprocess.PIPE, stderr=open(os.devnull, "w"))
        line = plugins.retryOnInterrupt(wmProc.stdout.readline)
        try:
            wmPid = int(line.strip())
            wmProc.stdout.close()
        except ValueError:
            sys.stderr.write("ERROR: Failed to start window manager")
            wmProc.stdout.close()
            return None, None
        return wmPid, wmProc

    def executableExists(self, machine, app, executable):
        retcode = app.runCommandOn(machine, ["which", executable], collectExitCode=True)
        return retcode == 0




"""
The various ways to launch the dynamic GUI from the static GUI
"""
from gi.repository import Gtk, GObject
import os
import sys
import stat
from texttestlib import plugins
from .. import guiplugins
from copy import copy, deepcopy
from io import StringIO

# Runs the dynamic GUI, but not necessarily with all the options available from the configuration


class BasicRunningAction:
    runNumber = 1

    def __init__(self, inputOptions):
        self.inputOptions = inputOptions
        self.testCount = 0

    def getTabTitle(self):
        return "Running"

    def messageAfterPerform(self):
        return self.performedDescription() + " " + self.describeTestsWithCount() + " at " + plugins.localtime() + "."

    def describeTestsWithCount(self):
        if self.testCount == 1:
            return "test " + self.getTestCaseSelection()[0].getRelPath()
        else:
            return str(self.testCount) + " tests"

    def performOnCurrent(self):
        self.startTextTestProcess(self.getUseCaseName(), ["-g"])

    def getTestsAffected(self, testSelOverride):
        if testSelOverride:
            return testSelOverride
        else:
            # Take a copy so we aren't fooled by selection changes
            return copy(self.currTestSelection)

    def getRunWriteDirectory(self, app):
        return os.path.join(self.getLogRootDirectory(app), "dynamic_run" + str(self.runNumber))

    def startTextTestProcess(self, usecase, runModeOptions, testSelOverride=None, filterFileOverride=None):
        app = self.getCurrentApplication()
        writeDir = self.getRunWriteDirectory(app)
        plugins.ensureDirectoryExists(writeDir)
        filterFile = self.createFilterFile(writeDir, filterFileOverride)
        ttOptions = runModeOptions + self.getTextTestOptions(filterFile, app, usecase)
        self.diag.info("Starting " + usecase + " run of TextTest with arguments " + repr(ttOptions))
        logFile = os.path.join(writeDir, "output.log")
        errFile = os.path.join(writeDir, "errors.log")
        BasicRunningAction.runNumber += 1
        description = "Dynamic GUI started at " + plugins.localtime()
        cmdArgs = self.getInterpreterArgs() + [sys.argv[0]] + ttOptions
        env = self.getNewUseCaseEnvironment(usecase)
        testsAffected = self.getTestsAffected(testSelOverride)
        guiplugins.processMonitor.startProcess(cmdArgs, description, env=env, killOnTermination=self.killOnTermination(),
                                               stdout=open(logFile, "w"), stderr=open(errFile, "w"),
                                               exitHandler=self.checkTestRun,
                                               exitHandlerArgs=(errFile, testsAffected, filterFile, usecase))

    def getCurrentApplication(self):
        return self.currAppSelection[0] if self.currAppSelection else self.validApps[0]

    def getLogRootDirectory(self, app):
        return app.writeDirectory

    def killOnTermination(self):
        return True

    def getNewUseCaseEnvironment(self, usecase):
        environ = deepcopy(os.environ)
        recScript = os.getenv("USECASE_RECORD_SCRIPT")
        if recScript:
            environ["USECASE_RECORD_SCRIPT"] = plugins.addLocalPrefix(recScript, usecase)
        repScript = os.getenv("USECASE_REPLAY_SCRIPT")
        if repScript:
            # Dynamic GUI might not record anything (it might fail) - don't try to replay files that
            # aren't there...
            dynRepScript = plugins.addLocalPrefix(repScript, usecase)
            if os.path.isfile(dynRepScript):
                environ["USECASE_REPLAY_SCRIPT"] = dynRepScript
            else:
                del environ["USECASE_REPLAY_SCRIPT"]
        return environ

    def getSignalsSent(self):
        return ["SaveSelection"]

    def createFilterFile(self, writeDir, filterFileOverride):
        # Because the description of the selection can be extremely long, we write it in a file and refer to it
        # This avoids too-long command lines which are a problem at least on Windows XP
        if filterFileOverride is None:
            filterFileName = os.path.join(writeDir, "gui_select")
            self.notify("SaveSelection", filterFileName)
            return filterFileName
        elif filterFileOverride is not NotImplemented:
            return filterFileOverride

    def getInterpreterArgs(self):
        if getattr(sys, 'frozen', False):
            return []
        interpreterArg = os.getenv("TEXTTEST_DYNAMIC_GUI_INTERPRETER", "") # Alternative interpreter for the dynamic GUI : mostly useful for coverage / testing
        if interpreterArg:
            return plugins.splitcmd(interpreterArg.replace("ttpython", sys.executable))
        else:  # pragma: no cover - cannot test without StoryText on dynamic GUI
            return [sys.executable]

    def getOptionGroups(self):
        return [self.optionGroup]

    def getTextTestOptions(self, filterFile, app, usecase):
        ttOptions = self.getCmdlineOptionForApps(filterFile)
        for group in self.getOptionGroups():
            ttOptions += self.getCommandLineArgs(group, self.getCommandLineKeys(usecase),
                                                 self.getCommandLineExcludeKeys())
        # May be slow to calculate for large test suites, cache it
        self.testCount = len(self.getTestCaseSelection())
        ttOptions += ["-count", str(self.testCount * self.getCountMultiplier())]
        if filterFile:
            ttOptions += ["-f", filterFile]
        tmpFilterDir = self.getTmpFilterDir(app)
        if tmpFilterDir:
            ttOptions += ["-fd", tmpFilterDir]
        return ttOptions

    def getCommandLineKeys(self, *args):
        # assume everything by default
        return []

    def getCommandLineExcludeKeys(self):
        return []

    def getCountMultiplier(self):
        return 1

    def getVanillaOption(self):
        options = []
        if "vanilla" in self.inputOptions:
            options.append("-vanilla")
            value = self.inputOptions.get("vanilla")
            if value:
                options.append(value)
        return options

    def getTmpFilterDir(self, app):
        return os.path.join(app.writeDirectory, "temporary_filter_files")

    def getAppIdentifier(self, app):
        return app.name + app.versionSuffix()

    def getCmdlineOptionForApps(self, filterFile):
        if not filterFile:
            return []

        apps = sorted(self.currAppSelection, key=self.validApps.index)
        appNames = list(map(self.getAppIdentifier, apps))
        return ["-a", ",".join(appNames)]

    def checkTestRun(self, errFile, testSel, filterFile, usecase):
        if not testSel:
            return
        if self.checkErrorFile(errFile, testSel, usecase):
            self.handleCompletion(testSel, filterFile, usecase)
            if len(self.currTestSelection) >= 1 and self.currTestSelection[0] in testSel:
                self.currTestSelection[0].filesChanged()

        testSel[0].notify("CloseDynamic", usecase)

    def checkErrorFile(self, errFile, testSel, usecase):
        if os.path.isfile(errFile) and len(testSel) > 0:
            errText = testSel[0].app.filterErrorText(errFile)
            if len(errText):
                self.notify("Status", usecase.capitalize() + " run failed for " + repr(testSel[0]))
                lines = errText.splitlines()
                maxLength = testSel[0].getConfigValue("lines_of_text_difference")
                if len(lines) > maxLength:
                    errText = "\n".join(lines[:maxLength])
                    errText += "\n(Very long message truncated. Full details can be seen in the file at\n" + errFile + ")"
                self.showErrorDialog(usecase.capitalize() + " run failed, with the following errors:\n" + errText)
                return False
        return True

    def handleCompletion(self, *args):
        pass  # only used when recording

    def getConfirmationMessage(self):
        # For extra speed we check the selection first before we calculate all the test cases again...
        if len(self.currTestSelection) > 1 or len(self.getTestCaseSelection()) > 1:
            multiTestWarning = self.getMultipleTestWarning()
            if multiTestWarning:
                return "You are trying to " + multiTestWarning + ".\nThis will mean lots of target application GUIs " + \
                       "popping up and may be hard to follow.\nAre you sure you want to do this?"
        return ""

    def getMultipleTestWarning(self):
        pass


class ReconnectToTests(BasicRunningAction, guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic)
        BasicRunningAction.__init__(self, inputOptions)
        self.addOption("v", "Version to reconnect to")
        self.addOption("reconnect", "Temporary result directory", os.getenv("TEXTTEST_TMP", ""), selectDir=True,
                       description="Specify a directory containing temporary texttest results. The reconnection will use a random subdirectory matching the version used.")
        appGroup = plugins.OptionGroup("Invisible")
        self.addApplicationOptions(allApps, appGroup, inputOptions)
        self.addSwitch("reconnfull", "Recomputation", options=appGroup.getOption("reconnfull").options)

    def _getStockId(self):
        return "connect"

    def _getTitle(self):
        return "Re_connect..."

    def getTooltip(self):
        return "Reconnect to previously run tests"

    def performedDescription(self):
        return "Reconnected to"

    def getUseCaseName(self):
        return "reconnect"

    def performOnCurrent(self):
        self.startTextTestProcess(self.getUseCaseName(), ["-g"] + self.getVanillaOption())

    def getAppIdentifier(self, app):
        # Don't send version data, we have our own field with that info and it has a slightly different meaning
        return app.name

    def getSizeAsWindowFraction(self):
        return 0.8, 0.7


class ReloadTests(BasicRunningAction, guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic)
        BasicRunningAction.__init__(self, inputOptions)
        self.appGroup = plugins.OptionGroup("Invisible")
        # We don't think reconnect can handle multiple roots currently
        # Can be a limitation of this for the moment.
        self.addApplicationOptions(allApps, self.appGroup, inputOptions)
        self.addSwitch("reconnfull", "Recomputation", options=self.appGroup.getOption("reconnfull").options)

    def getTmpDirectory(self):
        return self.currAppSelection[0].writeDirectory

    def _getStockId(self):
        return "connect"

    def _getTitle(self):
        return "Re_load tests..."

    def getTooltip(self):
        return "Reload current results into new dynamic GUI"

    def performedDescription(self):
        return "Reloaded"

    def getUseCaseName(self):
        return "reload"

    def performOnCurrent(self):
        if self.appGroup.getOptionValue("reconnfull") == 0:
            # We want to reload the results exactly as they are currently
            # This will only be possible if we make sure to save the teststate files first
            self.saveTestStates()
        self.startTextTestProcess(self.getUseCaseName(), [
                                  "-g", "-reconnect", self.getTmpDirectory()] + self.getVanillaOption())

    def saveTestStates(self):
        for test in self.currTestSelection:
            if test.state.isComplete():  # might look weird but this notification also comes in scripts etc.
                test.saveState()

    def getAppIdentifier(self, app):
        # Don't send version data, we have our own field with that info and it has a slightly different meaning
        return app.name


# base class for RunTests and RerunTests, i.e. all the options are available
class RunningAction(BasicRunningAction):
    originalVersion = ""

    def __init__(self, allApps, inputOptions):
        BasicRunningAction.__init__(self, inputOptions)
        self.optionGroups = []
        self.disablingInfo = {}
        self.disableWidgets = {}
        for groupName, disablingOption, disablingOptionValue in self.getGroupNames(allApps):
            group = plugins.OptionGroup(groupName)
            self.addApplicationOptions(allApps, group, inputOptions)
            self.optionGroups.append(group)
            if disablingOption:
                self.disablingInfo[self.getOption(disablingOption)] = disablingOptionValue, group

        self.temporaryGroup = plugins.OptionGroup("Temporary Settings")
        self.temporaryGroup.addOption("filetype", "File Type", "environment", possibleValues=self.getFileTypes(allApps))
        self.temporaryGroup.addOption("contents", "Contents", multilineEntry=True)

        RunningAction.originalVersion = self.getVersionString()

    def getFileTypes(self, allApps):
        ignoreTypes = ["testsuite", "knownbugs", "stdin", "input", "testcustomize.py"]
        fileTypes = []
        for app in allApps:
            for ft in app.defFileStems("builtin") + app.defFileStems("default"):
                if ft not in fileTypes and ft not in ignoreTypes:
                    fileTypes.append(ft)
        return fileTypes

    def getTextTestOptions(self, filterFile, app, *args):
        ret = BasicRunningAction.getTextTestOptions(self, filterFile, app, *args)
        contents = self.temporaryGroup.getValue("contents")
        if contents:
            fileType = self.temporaryGroup.getValue("filetype")
            writeDir = os.path.dirname(filterFile)
            tmpDir = self.makeTemporarySettingsDir(writeDir, app, fileType, contents)
            ret += ["-td", tmpDir]
        return ret

    def makeTemporarySettingsDir(self, writeDir, app, fileType, contents):
        tmpDir = os.path.join(writeDir, "temporary_settings")
        plugins.ensureDirectoryExists(tmpDir)
        fileName = os.path.join(tmpDir, fileType + "." + app.name + app.versionSuffix())
        with open(fileName, "w") as f:
            f.write(contents)
        return tmpDir

    def getGroupNames(self, allApps):
        if len(allApps) > 0:
            return allApps[0].getAllRunningGroupNames(allApps)
        else:
            configObject = self.makeDefaultConfigObject(self.inputOptions)
            return configObject.getAllRunningGroupNames(allApps)

    def createCheckBox(self, switch, *args):
        widget = guiplugins.OptionGroupGUI.createCheckBox(self, switch, *args)
        self.storeSwitch(switch, [widget])
        return widget

    def createRadioButtons(self, switch, *args):
        buttons = guiplugins.OptionGroupGUI.createRadioButtons(self, switch, *args)
        self.storeSwitch(switch, buttons)
        return buttons

    def storeSwitch(self, switch, widgets):
        if switch in self.disablingInfo:
            disablingOptionValue, group = self.disablingInfo[switch]
            if disablingOptionValue < len(widgets):
                self.disableWidgets[widgets[disablingOptionValue]] = switch, disablingOptionValue, group

    def getOption(self, optName):
        for group in self.optionGroups:
            opt = group.getOption(optName)
            if opt:
                return opt

    def getOptionGroups(self):
        return self.optionGroups

    def getCountMultiplier(self):
        return self.getCopyCount() * self.getVersionCount()

    def getCopyCount(self):
        return self.getOption("cp").getValue()

    def getVersionString(self):
        vOption = self.getOption("v")
        if vOption:
            versionString = vOption.getValue()
            return "" if versionString.startswith("<default>") else versionString
        else:
            return ""

    def getVersionCount(self):
        return self.getVersionString().count(",") + 1

    def performedDescription(self):
        timesToRun = self.getCopyCount()
        numberOfTests = self.testCount
        if timesToRun != 1:
            if numberOfTests > 1:
                return "Started " + str(timesToRun) + " copies each of"
            else:
                return "Started " + str(timesToRun) + " copies of"
        else:
            return "Started"

    def getMultipleTestWarning(self):
        app = self.currTestSelection[0].app
        for group in self.getOptionGroups():
            for switchName, desc in app.getInteractiveReplayOptions():
                if group.getSwitchValue(switchName, False):
                    return "run " + self.describeTests() + " with " + desc + " replay enabled"

    def getConfirmationMessage(self):
        runVersion = self.getVersionString()
        if self.originalVersion and self.originalVersion not in runVersion:
            return "You have tried to run a version ('" + runVersion + \
                   "') which is not based on the version you started with ('" + self.originalVersion + "').\n" + \
                   "This will result in an attempt to amalgamate the versions, i.e. to run version '" + \
                   self.originalVersion + "." + runVersion + "'.\n" + \
                   "If this isn't what you want, you will need to restart the static GUI with a different '-v' flag.\n\n" + \
                   "Are you sure you want to continue?"
        else:
            return BasicRunningAction.getConfirmationMessage(self)

    def createNotebook(self):
        notebook = Gtk.Notebook()
        notebook.set_name("sub-notebook for running")
        tabNames = ["Basic", "Advanced"]
        frames = []
        for group in self.optionGroups:
            if group.name in tabNames:
                label = Gtk.Label(label=group.name)
                tab = self.createTab(group, frames)
                notebook.append_page(tab, label)
            elif len(list(group.keys())) > 0:
                frames.append(self.createFrame(group, group.name))
        self.connectDisablingSwitches()
        notebook.show_all()
        self.widget = notebook
        return notebook

    def createTab(self, group, frames):
        tabBox = Gtk.VBox()
        if frames:
            frames.append(self.createFrame(group, "Miscellaneous"))
            frames.append(self.createFrame(self.temporaryGroup,  self.temporaryGroup.name))
            for frame in frames:
                tabBox.pack_start(frame, False, False, 8)
        else:
            self.fillVBox(tabBox, group)
        if isinstance(self, guiplugins.ActionTabGUI):
            # In a tab, we need to duplicate the buttons for each subtab
            # In a dialog we should not do this
            self.createButtons(tabBox)
        widget = self.addScrollBars(tabBox, hpolicy=Gtk.PolicyType.AUTOMATIC)
        widget.set_name(group.name + " Tab")
        return widget

    def updateSensitivity(self, widget, data):
        switch, disablingOptionValue, group = data
        sensitive = switch.getValue() != disablingOptionValue
        self.setGroupSensitivity(group, sensitive, ignoreWidget=widget)

    def connectDisablingSwitches(self):
        for widget, data in list(self.disableWidgets.items()):
            self.updateSensitivity(widget, data)
            widget.connect("toggled", self.updateSensitivity, data)
        self.disableWidgets = {}  # not needed any more

    def notifyReset(self, *args):
        for optionGroup in self.optionGroups:
            optionGroup.reset()
        self.temporaryGroup.reset()

    def _getStockId(self):
        return "media-play"


class RunTests(RunningAction, guiplugins.ActionTabGUI):
    def __init__(self, allApps, dummy, inputOptions):
        guiplugins.ActionTabGUI.__init__(self, allApps)
        RunningAction.__init__(self, allApps, inputOptions)

    def _getTitle(self):
        return "_Run"

    def getTooltip(self):
        return "Run selected tests"

    def getUseCaseName(self):
        return "dynamic"

    def createView(self):
        return self.createNotebook()

    def updateName(self, nameOption, name):
        if name:
            nameOption.setValue("Tests started from " + repr(name) + " at <time>")

    def notifySetRunName(self, name):
        nameOption = self.getOption("name")
        self.updateName(nameOption, name)

    def addApplicationOptions(self, allApps, group, inputOptions):
        guiplugins.ActionTabGUI.addApplicationOptions(self, allApps, group, inputOptions)
        nameOption = group.getOption("name")
        if nameOption:
            self.updateName(nameOption, nameOption.getValue())


class RerunTests(RunningAction, guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dummy, inputOptions):
        guiplugins.ActionDialogGUI.__init__(self, allApps)
        RunningAction.__init__(self, allApps, inputOptions)
        self.rerunTests = []
        for group in self.optionGroups:
            if group.name == "Basic":
                group.addSwitch("mark", "Mark test as rerun in current window", 1, description="Test will be automatically marked as having a newer result elsewhere." +
                                "It will also be possible to import this result here in future.")
                return

    def updateOptions(self):
        checkouts = []
        for app in self.currAppSelection:
            path = app.getGivenCheckoutPath(app)
            if path and path not in checkouts and not app.isAutogeneratedFromCwd(path):
                checkouts.append(path)
        if checkouts:
            self.getOption("c").setValue(",".join(checkouts))

    def _getTitle(self):
        return "_Rerun"

    def getTooltip(self):
        return "Rerun selected tests"

    def getUseCaseName(self):
        return "rerun"

    def getSignalsSent(self):
        return RunningAction.getSignalsSent(self) + ["Mark"]

    def killOnTermination(self):
        return False  # don't want rerun GUIs to disturb each other like this

    def getCommandLineExcludeKeys(self):
        return ["mark"]

    def getTmpFilterDir(self, app):
        return ""  # don't want selections returned here, send them to the static GUI

    def getConfirmationMessage(self):
        return BasicRunningAction.getConfirmationMessage(self)

    def getTextTestOptions(self, filterFile, app, usecase):
        rerunId = str(self.runNumber)
        if not self.getOption("name").getValue():
            rerunId += " from " + plugins.startTimeString()
        return RunningAction.getTextTestOptions(self, filterFile, app, usecase) + ["-rerun", rerunId]

    def getLogRootDirectory(self, app):
        if "f" in self.inputOptions:
            logRootDir = os.path.dirname(self.inputOptions["f"])
            if os.path.basename(logRootDir).startswith("dynamic_run"):
                return logRootDir
        return BasicRunningAction.getLogRootDirectory(self, app)

    def getRunWriteDirectory(self, app):
        fromDynamic = self.startedFromDynamicGui(app)
        if fromDynamic:
            return os.path.join(os.getenv("TEXTTEST_TMP"), "dynamic_run" + str(self.runNumber) + "_" + plugins.startTimeString().replace(":", ""))
        else:
            return RunningAction.getRunWriteDirectory(self, app)

    def startedFromDynamicGui(self, app):
        return not app.useStaticGUI() and not "dynamic_run" in os.path.basename(self.getLogRootDirectory(app))

    def getExtraParent(self, app):
        for other in self.validApps:
            if app in other.extras:
                return other

    def getExtraVersions(self, app):
        if app.extras or any((v.startswith("copy_") for v in app.versions)):
            return []
        extraParent = self.getExtraParent(app)
        if extraParent:
            return [v for v in app.versions if v not in extraParent.versions]
        else:
            extrasGiven = app.getConfigValue("extra_version")
            return [v for v in app.versions if v in extrasGiven]

    def getAppIdentifier(self, app):
        parts = [app.name] + self.getExtraVersions(app)
        return ".".join(parts)

    def checkTestRun(self, errFile, testSel, filterFile, usecase):
        # Don't do anything with the files, but do produce popups on failures and notify when complete
        self.checkErrorFile(errFile, testSel, usecase)
        if len(testSel) > 0:
            app = testSel[0].app
            if not app.keepTemporaryDirectories() and self.startedFromDynamicGui(app):
                writeDir = os.path.dirname(errFile)
                plugins.rmtree(writeDir)
            testSel[0].notify("CloseDynamic", usecase)

    def getOrderedOptions(self, optionGroup):
        options, switches = self.extractSwitches(optionGroup)
        return options + switches

    def fillVBox(self, vbox, optionGroup):
        if optionGroup is self.optionGroup:
            notebook = self.createNotebook()
            vbox.pack_start(notebook, True, True, 0)
            return None, None  # no file chooser info
        else:
            return guiplugins.ActionDialogGUI.fillVBox(self, vbox, optionGroup, includeOverrides=False)

    def performOnCurrent(self):
        RunningAction.performOnCurrent(self)
        if self.getOption("mark").getValue():
            lastRunText = str(self.runNumber - 1)
            self.rerunTests = copy(self.currTestSelection)
            for test in self.currTestSelection:
                self.notify("Mark", test, "Test is being rerun in another window, numbered " +
                            lastRunText, "Rerun " + lastRunText)

    def getTestCaseSelection(self):
        if self.rerunTests:
            tests = self.rerunTests
            self.rerunTests = []
            return tests
        else:
            return guiplugins.ActionDialogGUI.getTestCaseSelection(self)

    def getSizeAsWindowFraction(self):
        return 0.8, 0.9


class RecordTest(BasicRunningAction, guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic)
        BasicRunningAction.__init__(self, inputOptions)
        self.recordTime = None
        self.currentApp = None
        if len(allApps) > 0:
            self.currentApp = allApps[0]
        self.addOptions()
        self.addSwitches()

    def addOptions(self):
        defaultVersion, defaultCheckout = "", ""
        if self.currentApp:
            defaultVersion = self.currentApp.getFullVersion()
            defaultCheckout = self.currentApp.getCheckoutForDisplay()

        self.addOption("v", "Version to record", defaultVersion)
        self.addOption("c", self.getCheckoutLabel(), defaultCheckout)
        self.addOption("m", "Record on machine")

    def getCheckoutLabel(self):
        # Sometimes configurations might want to use their own term in place of "checkout"
        return "Checkout to use for recording"

    def addSwitches(self):
        if self.currentApp and self.currentApp.usesCaptureMock():
            self.currentApp.addCaptureMockSwitch(self.optionGroup, value=1)  # record new by default
        self.addSwitch("rep", "Automatically replay test after recording it", 2,
                       options=["Disabled", "In background", "Using dynamic GUI"])
        if self.currentApp and self.currentApp.getConfigValue("extra_test_process_postfix"):
            self.addSwitch("mult", "Record multiple runs of system")

    def correctTestClass(self):
        return "test-case"

    def _getStockId(self):
        return "media-record"

    def messageAfterPerform(self):
        return "Started record session for " + self.describeTests()

    def touchFiles(self, test):
        for postfix in test.getConfigValue("extra_test_process_postfix"):
            if not test.getFileName("usecase" + postfix):
                fileName = os.path.join(test.getDirectory(), "usecase" + postfix + "." + test.app.name)
                with open(fileName, "w") as f:
                    f.write("Dummy file to indicate we should record multiple runs\n")

    def performOnCurrent(self):
        test = self.currTestSelection[0]
        if self.optionGroup.getSwitchValue("mult"):
            self.touchFiles(test)
        self.updateRecordTime(test)
        self.startTextTestProcess("record", ["-g", "-record"] + self.getVanillaOption())

    def shouldShowCurrent(self, *args):
        # override the default so it's disabled if there are no apps
        return len(self.validApps) > 0 and guiplugins.ActionDialogGUI.shouldShowCurrent(self, *args)

    def isValidForApp(self, app):
        return app.getConfigValue("use_case_record_mode") != "disabled" and \
            app.getConfigValue("use_case_recorder") != "none"

    def updateOptions(self):
        if self.currentApp is not self.currAppSelection[0]:
            self.currentApp = self.currAppSelection[0]
            self.optionGroup.setOptionValue("v", self.currentApp.getFullVersion())
            self.optionGroup.setOptionValue("c", self.currentApp.getCheckoutForDisplay())
            return True
        else:
            return False

    def getUseCaseFile(self, test):
        return test.getFileName("usecase", self.optionGroup.getOptionValue("v"))

    def updateRecordTime(self, test):
        file = self.getUseCaseFile(test)
        if file:
            self._updateRecordTime(file)

    def _updateRecordTime(self, file):
        newTime = plugins.modifiedTime(file)
        if newTime != self.recordTime:
            self.recordTime = newTime
            return True
        else:
            return False

    def getChangedUseCaseVersion(self, test):
        test.refreshFiles()  # update cache after record run
        file = self.getUseCaseFile(test)
        if not file or not self._updateRecordTime(file):
            return

        parts = os.path.basename(file).split(".")
        return ".".join(parts[2:])

    def getMultipleTestWarning(self):
        return "record " + self.describeTests() + " simultaneously"

    def handleCompletion(self, testSel, filterFile, usecase):
        test = testSel[0]
        if usecase == "record":
            changedUseCaseVersion = self.getChangedUseCaseVersion(test)
            replay = self.optionGroup.getSwitchValue("rep")
            if changedUseCaseVersion is not None and replay:
                replayOptions = self.getVanillaOption() + self.getReplayRunModeOptions(changedUseCaseVersion)
                self.startTextTestProcess("replay", replayOptions, testSel, filterFile)
                message = "Recording completed for " + repr(test) + \
                          ". Auto-replay of test now started. Don't submit the test manually!"
                self.notify("Status", message)
            else:
                self.notify("Status", "Recording completed for " + repr(test) + ", not auto-replaying")
        else:
            self.notify("Status", "Recording and auto-replay completed for " + repr(test))

    def getCommandLineKeys(self, usecase):
        keys = ["v", "c", "m"]
        if usecase == "record":
            keys.append("rectraffic")
        return keys

    def getReplayRunModeOptions(self, overwriteVersion):
        if self.optionGroup.getSwitchValue("rep") == 2:
            return ["-autoreplay", "-g"]
        else:
            return ["-autoreplay", "-o", overwriteVersion]

    def _getTitle(self):
        return "Record _Use-Case"

    def getSizeAsWindowFraction(self):
        return 0.5, 0.5


class RunScriptAction(BasicRunningAction):
    def getUseCaseName(self):
        return "script"

    def performOnCurrent(self, **kw):
        self.startTextTestProcess(self.getUseCaseName(), ["-g"] + self.getVanillaOption(), **kw)

    def getCommandLineArgs(self, optionGroup, *args):
        args = [self.scriptName()]
        for key, option in list(optionGroup.options.items()):
            args.append(key + "=" + str(option.getValue()))

        return ["-s", " ".join(args)]


class ReplaceText(RunScriptAction, guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        RunScriptAction.__init__(self, inputOptions)
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic)
        self.addSwitch("regexp", "Enable regular expressions", 1)
        self.addOption("old", "Text or regular expression to search for", multilineEntry=True)
        self.addOption("new", "Text to replace it with (may contain regexp back references)", multilineEntry=True)
        self.addOption("file", "File stem(s) to perform replacement in", allocateNofValues=2)
        self.storytextDirs = {}

    def getCmdlineOptionForApps(self, filterFile):
        options = RunScriptAction.getCmdlineOptionForApps(self, filterFile)
        if self.shouldAddShortcuts():
            options[1] = options[1] + ",shortcut"
            directoryStr = os.path.dirname(filterFile) + os.pathsep + os.pathsep.join(self.inputOptions.rootDirectories)
            options += ["-d", directoryStr]
        return options

    def createFilterFile(self, writeDir, filterFileOverride):
        filterFileName = RunScriptAction.createFilterFile(self, writeDir, filterFileOverride)
        if self.shouldAddShortcuts():
            storytextDir = self.storytextDirs[self.currAppSelection[0]]
            self.createShortcutApps(writeDir)
            with open(filterFileName, "a") as filterFile:
                filterFile.write("appdata=shortcut\n")
                filterFile.write(os.path.basename(storytextDir) + "\n")
        return filterFileName

    def notifyAllStems(self, allStems, defaultTestFile):
        self.optionGroup.setValue("file", defaultTestFile)
        self.optionGroup.setPossibleValues("file", allStems)

    def notifyNewTestSelection(self, *args):
        guiplugins.ActionDialogGUI.notifyNewTestSelection(self, *args)
        if len(self.storytextDirs) > 0:
            self.addSwitch("includeShortcuts", "Replace text in shortcut files", 0)

    def shouldAddShortcuts(self):
        return len(self.storytextDirs) > 0 and self.optionGroup.getOptionValue("includeShortcuts") > 0

    def notifyUsecaseRename(self, argstr, *args):
        self.showQueryDialog(self.getParentWindow(), "Usecase names were renamed. Would you like to update them in all usecases now?",
                             Gtk.STOCK_DIALOG_WARNING, "Confirmation", self.respondUsecaseRename, respondData=(argstr, False, "*usecase*,stdout"))

    def notifyShortcutRename(self, argstr, *args):
        self.showQueryDialog(self.getParentWindow(), "Shortcuts were renamed. Would you like to update all usecases now?",
                             Gtk.STOCK_DIALOG_WARNING, "Confirmation", self.respondUsecaseRename, respondData=(argstr, True, "*usecase*"))

    def notifyShortcutRemove(self, argstr, *args):
        self.showQueryDialog(self.getParentWindow(), "Shortcuts were removed. Would you like to update all usecases now?",
                             Gtk.STOCK_DIALOG_WARNING, "Confirmation", self.respondUsecaseRename, respondData=(argstr, True, "*usecase*"))

    def respondUsecaseRename(self, dialog, ans, args):
        if ans == Gtk.ResponseType.YES:
            oldName, newName = args[0].split(" renamed to ")
            if args[1]:
                self.optionGroup.setValue("regexp", 1)
                self.addSwitch("argsReplacement", "", 1)
            self.optionGroup.setValue("file", args[2])
            self.optionGroup.setValue("old", oldName.strip("'"))
            self.optionGroup.setValue("new", newName.strip("'"))
            self.performOnCurrent(filterFileOverride=NotImplemented)
        dialog.hide()

    def createShortcutApps(self, writeDir):
        for app, storyTextHome in list(self.storytextDirs.items()):
            self.createConfigFile(app, writeDir)
            self.createTestSuiteFile(app, storyTextHome, writeDir)

    def createConfigFile(self, app, writeDir):
        configFileName = os.path.join(writeDir, "config.shortcut")
        with open(configFileName, "w") as configFile:
            configFile.write("executable:None\n")
            configFile.write("filename_convention_scheme:standard\n")
            configFile.write("use_case_record_mode:GUI\n")
            configFile.write("use_case_recorder:storytext")
        return configFileName

    def createTestSuiteFile(self, app, storyTextHome, writeDir):
        suiteFileName = os.path.join(writeDir, "testsuite.shortcut")
        with open(suiteFileName, "w") as suiteFile:
            suiteFile.write(storyTextHome + "\n")
        return suiteFileName

    def scriptName(self):
        return "default.ReplaceText"

    def _getTitle(self):
        return "Replace Text in Files"

    def getTooltip(self):
        return "Replace text in multiple test files"

    def performedDescription(self):
        return "Replaced text in files for"

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.5, 0.5

    def notifyUsecaseHome(self, suite, usecaseHome):
        self.storytextDirs[suite.app] = usecaseHome

    def _respond(self, saidOK=True, dialog=None, fileChooserOption=None):
        if saidOK and not self.optionGroup.getValue("old"):
            self.showWarningDialog("Text or regular expression to search for cannot be empty")
        else:
            guiplugins.ActionDialogGUI._respond(self, saidOK, dialog, fileChooserOption)


class TestFileFilterHelper:
    def showTextInFile(self, text):
        test = self.currTestSelection[0]
        fileName = self.currFileSelection[0][0]
        root = test.getEnvironment("TEXTTEST_SANDBOX_ROOT")
        plugins.ensureDirectoryExists(root)
        tmpFileNameLocal = os.path.basename(fileName) + " (FILTERED)"
        tmpFileName = os.path.join(root, tmpFileNameLocal)
        bakFileName = tmpFileName + ".bak"
        if os.path.isfile(bakFileName):
            os.remove(bakFileName)
        if os.path.isfile(tmpFileName):
            os.rename(tmpFileName, bakFileName)
        with open(tmpFileName, "w") as tmpFile:
            tmpFile.write(text)

        # Don't want people editing by mistake, remove write permissions
        os.chmod(tmpFileName, stat.S_IREAD)
        self.notify("ViewReadonlyFile", tmpFileName)

    def getSignalsSent(self):
        return ["ViewReadonlyFile"]


class TestFileFiltering(TestFileFilterHelper, guiplugins.ActionGUI):
    def _getTitle(self):
        return "Test Filtering"

    def isActiveOnCurrent(self, *args):
        return guiplugins.ActionGUI.isActiveOnCurrent(self) and len(self.currFileSelection) == 1

    def getTextToShow(self, test, fileName):
        return test.app.applyFiltering(test, fileName)

    def performOnCurrent(self):
        self.reloadConfigForSelected()  # Always make sure we're up to date here
        text = self.getTextToShow(self.currTestSelection[0], self.currFileSelection[0][0])
        self.showTextInFile(text)


class ShowFilters(TestFileFilterHelper, guiplugins.ActionResultDialogGUI):
    def __init__(self, *args, **kw):
        guiplugins.ActionResultDialogGUI.__init__(self, *args, **kw)
        self.textBuffer = None
        self.filtersWithModels = []
        self.toRemove = {}

    def _getTitle(self):
        return "Show Filters"

    def isActiveOnCurrent(self):
        return len(self.currFileSelection) == 1 and len(self.currTestSelection) == 1

    def addContents(self):
        self.reloadConfigForSelected()  # Always make sure we're up to date here
        fileName = self.currFileSelection[0][0]
        test = self.currTestSelection[0]
        allFilters, versionApp = test.app.getAllFilters(test, fileName)
        if allFilters:
            self.addFilterBoxes(allFilters, fileName, test, versionApp)
        else:
            messageBox = self.createDialogMessage(
                "No run_dependent_text filters defined for file '" + os.path.basename(fileName) + "' for this test.", Gtk.STOCK_DIALOG_INFO)
            self.dialog.vbox.pack_start(messageBox, True, True, 0)

    def editFilter(self, cell, path, newText, model):
        lineFilter = model[path][0]
        model[path][0] = lineFilter.makeNew(newText)

    def showToggled(self, cell, path, model):
        # Toggle the toggle button
        newValue = not model[path][1]
        model[path][1] = newValue

    def setText(self, column, cell, model, iter):
        cell.set_property('text', model.get_value(iter, 0).originalText)

    def getStem(self, fileName):
        return os.path.basename(fileName).split(".")[0]

    def addRow(self, menuItem, model, selection, configKey):
        iter = self.getSelectedIters(selection)[-1]
        lineFilter, relpath, stem = model.get(iter, 0, 2, 3)
        newIter = model.insert_after(iter, [lineFilter.makeNew(""), True, relpath, stem, ""])
        path = model.get_path(newIter)
        column = selection.get_tree_view().get_column(0)
        selection.get_tree_view().set_cursor(path, column, start_editing=True)

    def getSelectedIters(self, selection):
        iters = []

        def addSelIter(model, path, iter):
            iters.append(iter)

        selection.selected_foreach(addSelIter)
        return iters

    def removeRow(self, menuItem, model, selection, configKey):
        for iter in self.getSelectedIters(selection):
            self.addChangeData(model, iter, configKey, self.toRemove)
            model.remove(iter)

    def makePopup(self, *args):
        menu = Gtk.Menu()
        menuItem = Gtk.MenuItem("Add Row")
        menu.append(menuItem)
        menuItem.connect("activate", self.addRow, *args)
        menuItem.show()
        menuItem = Gtk.MenuItem("Remove")
        menu.append(menuItem)
        menuItem.connect("activate", self.removeRow, *args)
        menuItem.show()
        return menu

    def showPopup(self, treeview, event, popupMenu):
        if event.button == 3:
            pathInfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            if pathInfo is not None:
                treeview.grab_focus()
                popupMenu.popup(None, None, None, event.button, event.time)

    def addFilterBoxes(self, allFilters, fileName, test, versionApp):
        filterFrame = Gtk.Frame.new("Filters to apply")
        filterFrame.set_border_width(1)
        vbox = Gtk.VBox()
        for filterObj in allFilters:
            listStore = Gtk.ListStore(GObject.TYPE_PYOBJECT, bool, str, str, str)
            for lineFilter in filterObj.lineFilters:
                lineFilterFile, stem = test.getConfigFileDefining(
                    versionApp, filterObj.configKey, self.getStem(fileName), lineFilter.originalText)
                relPath = plugins.relpath(lineFilterFile, test.app.getDirectory()) if lineFilterFile else "???"
                listStore.append([lineFilter, True, relPath, stem, lineFilter.originalText])
            treeView = Gtk.TreeView(listStore)
            treeView.set_name(filterObj.configKey + " Tree View")

            cell = Gtk.CellRendererText()
            cell.set_property('editable', True)
            cell.connect('edited', self.editFilter, listStore)
            column = Gtk.TreeViewColumn(filterObj.configKey.replace("_", "__"))
            column.pack_start(cell, True)
            column.set_cell_data_func(cell, self.setText)
            treeView.append_column(column)

            toggleCell = Gtk.CellRendererToggle()
            toggleCell.set_property('activatable', True)
            toggleCell.connect("toggled", self.showToggled, listStore)
            column = Gtk.TreeViewColumn("Enabled", toggleCell, active=1)
            treeView.append_column(column)

            cell = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn("Config File", cell, text=2)
            treeView.append_column(column)

            selection = treeView.get_selection()
            selection.set_mode(Gtk.SelectionMode.MULTIPLE)

            popup = self.makePopup(listStore, selection, filterObj.configKey)
            treeView.connect("button_press_event", self.showPopup, popup)

            frame = Gtk.Frame()
            frame.set_border_width(1)
            frame.add(treeView)
            vbox.pack_start(frame, True, True, 0)
            self.filtersWithModels.append((filterObj, listStore))
        filterFrame.add(vbox)
        self.dialog.vbox.pack_start(filterFrame, False, True, 0)
        self.dialog.vbox.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, True, 0)
        frame, self.textBuffer = self.createTextWidget("Filter Text View", scroll=True)
        frame.set_label("Text to filter (copied from " + os.path.basename(fileName) + ")")
        with open(fileName) as f:
            self.textBuffer.set_text(f.read())
        self.dialog.vbox.pack_start(frame, True, True, 0)

    def testFiltering(self, *args):
        filteredText = self.getFilteredText()
        self.showTextInFile(filteredText)

    def getText(self):
        return self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), True)

    def updateFilter(self, fileFilter, model):
        enabledFilters = []

        def addFilter(model, path, iter):
            if model.get_value(iter, 1):
                enabledFilters.append(model.get_value(iter, 0))
        model.foreach(addFilter)
        fileFilter.lineFilters = enabledFilters

    def getFilteredText(self):
        inFile = StringIO()
        inFile.write(self.getText())
        inFile.seek(0)
        for fileFilter, model in self.filtersWithModels:
            self.updateFilter(fileFilter, model)
            outFile = StringIO()
            fileFilter.filterFile(inFile, outFile)
            inFile.close()
            inFile = outFile
            inFile.seek(0)
        value = outFile.getvalue()
        outFile.close()
        return value

    def addChangeData(self, model, iter, configKey, changes):
        fileName, stem, origText = model.get(iter, 2, 3, 4)
        oldLine = stem + ":" + origText
        newLine = None if changes is self.toRemove else stem + ":" + model.get_value(iter, 0).originalText
        changes.setdefault(fileName, {}).setdefault(configKey, []).append((oldLine, newLine))

    def getChanges(self):
        changes = {}

        def addChange(model, path, iter, configKey):
            text = model.get_value(iter, 0).originalText
            oldText = model.get_value(iter, 4)
            if text != oldText:
                self.addChangeData(model, iter, configKey, changes)
        for fileFilter, model in self.filtersWithModels:
            model.foreach(addChange, fileFilter.configKey)
        for fileName, removeData in list(self.toRemove.items()):
            for removeKey, removeLines in list(removeData.items()):
                changes.setdefault(fileName, {}).setdefault(removeKey, []).extend(removeLines)
        self.toRemove = {}
        return changes

    def getSectionName(self, line):
        pos = line.find("]")
        return line[1:pos]

    def applyChanges(self, line, changeList):
        for oldText, newText in changeList:
            if line.startswith(oldText):
                if oldText.endswith(":") and newText.startswith(oldText):
                    # Adding a new row
                    return line, newText
                else:
                    return line.replace(oldText, newText, 1) if newText else None, None
        return line, None

    def saveChanges(self):
        changesByFile = self.getChanges()
        if changesByFile:
            app = self.currAppSelection[0]
            for fileName, changes in list(changesByFile.items()):
                newFileLines = []
                fullPath = os.path.join(app.getDirectory(), fileName)
                currSection = None
                currNewLines = []
                with open(fullPath) as f:
                    for line in f:
                        if line.startswith("["):
                            currSection = self.getSectionName(line)
                        elif currSection in changes:
                            line, newLine = self.applyChanges(line, changes.get(currSection))
                            if newLine:
                                currNewLines.append(newLine)
                        if line is not None:
                            if currNewLines and not line.startswith(currNewLines[0].split(":")[0]):
                                for newLine in currNewLines:
                                    newFileLines.append(newLine + "\n")
                                currNewLines = []
                            newFileLines.append(line)

                with open(fullPath, "w") as f:
                    for line in newFileLines:
                        f.write(line)

    def createButtons(self):
        button = self.dialog.add_button('Test Filtering', Gtk.ResponseType.NONE)
        button.connect("clicked", self.testFiltering)
        self.dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.dialog.add_button("Apply and Close", Gtk.ResponseType.ACCEPT)
        self.dialog.set_default_response(Gtk.ResponseType.ACCEPT)
        self.dialog.connect("response", self.respond)

    def respond(self, dialog, responseId):
        if responseId == Gtk.ResponseType.ACCEPT:
            self.saveChanges()
        guiplugins.ActionResultDialogGUI.respond(self, dialog, responseId)

    def getSizeAsWindowFraction(self):
        return 0.8, 0.9


class InsertShortcuts(RunScriptAction, guiplugins.OptionGroupGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.OptionGroupGUI.__init__(self, allApps, dynamic)
        RunScriptAction.__init__(self, inputOptions)

    def scriptName(self):
        return "default.InsertShortcuts"

    def _getTitle(self):
        return "Insert Shortcuts into Usecases"

    def getTootip(self):
        return self._getTitle()

    def notifyShortcut(self, *args):
        self.showQueryDialog(self.getParentWindow(), "New shortcuts were created. Would you like to insert them into all usecases now?",
                             Gtk.STOCK_DIALOG_WARNING, "Confirmation", self.respondShortcut)

    def respondShortcut(self, dialog, ans, *args):
        if ans == Gtk.ResponseType.YES:
            self.performOnCurrent(filterFileOverride=NotImplemented)
        dialog.hide()

    def performedDescription(self):
        return "Inserted shortcuts into usecases for"

    def isValidForApp(self, app):
        return app.getConfigValue("use_case_record_mode") != "disabled" and \
            app.getConfigValue("use_case_recorder") != "none"


def getInteractiveActionClasses(dynamic):
    if dynamic:
        return [RerunTests, ReloadTests]
    else:
        return [RunTests, RecordTest, ReconnectToTests, ReplaceText, ShowFilters, TestFileFiltering, InsertShortcuts]



from gi.repository import Gtk, Gdk

# We register our own stock items here.
# The code is inspired by the example
# in the pygtk FAQ, item 9.5.


def register(window):
    # needs review, StockItems are deprecated MB 2018-12-05
    loadItem = Gtk.StockItem()
    loadItem.stock_id = 'texttest-stock-load'
    loadItem.label = '_Load'
    loadItem.modifier = Gdk.ModifierType.CONTROL_MASK
    loadItem.keyval = Gdk.keyval_from_name('L')
    creditsItem = Gtk.StockItem()
    creditsItem.stock_id = 'texttest-stock-credits'
    creditsItem.label = 'C_redits'

    # We're too lazy to make our own icons,
    # so we use regular stock icons.
    aliases = [('texttest-stock-load', Gtk.STOCK_OPEN),
               ('texttest-stock-credits', Gtk.STOCK_ABOUT)]

    Gtk.stock_add([loadItem, creditsItem])
    factory = Gtk.IconFactory()
    factory.add_default()
    style = window.get_style()
    for new_stock, alias in aliases:
        icon_set = style.lookup_icon_set(alias)
        factory.add(new_stock, icon_set)
